# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.lines import Line2D
import mpl_toolkits.mplot3d as mp3d

from .batch_generation import create_batch_rw


def visualize_random_walk(RW, color=True, colorbar=True):
    dim = len(set(RW.columns).intersection({'x', 'y', 'z'}))
    if dim == 1:
        fig, ax = plt.subplots()
        if color:
            points = np.array([RW.t, RW.x]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            norm = plt.Normalize(RW.t.min(), RW.t.max())
            lc = LineCollection(segments, cmap='viridis', norm=norm)
            lc.set_array(RW.t)
            line = ax.add_collection(lc)
            ax.axis('square')
            if colorbar:
                cbar = plt.colorbar(line, ax=ax)
                cbar.set_label('time')
        else:
            line = ax.plot(RW.x)
    elif dim == 2:
        fig, ax = plt.subplots()
        if color:
            points = np.array([RW.x, RW.y]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            norm = plt.Normalize(RW.t.min(), RW.t.max())
            lc = LineCollection(segments, cmap='viridis', norm=norm)
            lc.set_array(RW.t)
            line = ax.add_collection(lc)
            ax.axis('square')
            if colorbar:
                cbar = plt.colorbar(line, ax=ax)
                cbar.set_label('time')
        else:
            line = ax.plot(RW.x, RW.y)
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        if color:
            points = np.array([RW.x, RW.y, RW.z]).T.reshape(-1, 1, 3)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            norm = plt.Normalize(RW.t.min(), RW.t.max())
            lc = mp3d.art3d.Line3DCollection(segments, cmap='viridis',
                                             norm=norm)
            lc.set_array(RW.t)
            line = ax.add_collection(lc)
            rmin, rmax = np.min(points), np.max(points)
            ax.set_xlim(rmin, rmax)
            ax.set_ylim(rmin, rmax)
            ax.set_zlim(rmin, rmax)
            if colorbar:
                cbar = plt.colorbar(line, ax=ax)
                cbar.set_label('time')
        else:
            ax.plot(RW.x, RW.y, RW.z)
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')


def visualize_rw_hmm(RW):
    dim = len(set(RW.columns).intersection({'x', 'y', 'z'}))
    nstates = RW.state.max()
    dict_state_color = {}
    for k in range(int(nstates)+1):
        dict_state_color[k] = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w'][k]

    if dim == 1:
        fig, ax = plt.subplots()
        points = np.array([RW.t, RW.x]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        norm = plt.Normalize(0, nstates)
        lc = LineCollection(segments,
                            colors=[dict_state_color[x] for x in RW.state])
        line = ax.add_collection(lc)
        ax.axis('square')
        manual_legend = [Line2D([0], [0], marker='_', markersize=6, color=v,
                                label=f'diffusion type {k+1}')
                         for k, v in dict_state_color.items()]
        ax.legend(handles=manual_legend, loc=0)
    elif dim == 2:
        fig, ax = plt.subplots()
        points = np.array([RW.x, RW.y]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        norm = plt.Normalize(0, nstates)
        lc = LineCollection(segments,
                            colors=[dict_state_color[x] for x in RW.state])
        line = ax.add_collection(lc)
        ax.axis('square')
        manual_legend = [Line2D([0], [0], marker='_', markersize=6, color=v,
                                label=f'diffusion type {k+1}')
                         for k, v in dict_state_color.items()]
        ax.legend(handles=manual_legend, loc=0)
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        points = np.array([RW.x, RW.y, RW.z]).T.reshape(-1, 1, 3)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        norm = plt.Normalize(0, nstates)
        lc = mp3d.art3d.Line3DCollection(segments,
                                         colors=[dict_state_color[x]
                                                 for x in RW.state])
        line = ax.add_collection(lc)
        ax.axis('square')
        manual_legend = [Line2D([0], [0], marker='_', markersize=6, color=v,
                                label=f'diffusion type {k+1}')
                         for k, v in dict_state_color.items()]
        ax.legend(handles=manual_legend, loc=0)
        rmin, rmax = np.min(points), np.max(points)
        ax.set_xlim(rmin, rmax)
        ax.set_ylim(rmin, rmax)
        ax.set_zlim(rmin, rmax)


def plot_convex_hull(XY, hull, imax):
    """Plots a convex hull
    """
    plt.figure()
    plt.plot(XY[:, 0], XY[:, 1], linewidth=0.5)
    for simplex in hull.simplices[:-1]:
        plt.plot(XY[simplex, 0], XY[simplex, 1], 'r-')
    plt.plot(XY[hull.simplices[-1], 0],
             XY[hull.simplices[-1], 1], 'r-', label='Convex hull')
    plt.scatter(XY[hull.vertices, 0], XY[hull.vertices, 1], s=10,
                color='red', label='Nodes of the convex hull')
    # max_distance
    n = len(hull.vertices)
    p1, p2 = imax//n, imax % n
    plt.plot(XY[hull.vertices][[p1, p2], 0], XY[hull.vertices]
             [[p1, p2], 1], 'r--', label='max distance')
    plt.legend()
    plt.xlabel('x')
    plt.ylabel('y')
    plt.axis('square')


def plot_embedding_with_color(mu, logvar, color, plot_every=10, name=None,
                              square=False, s=0.1, alpha=1, figsize=(9, 6)):
    fig, axs = plt.subplots(figsize=figsize, ncols=2)
    axs[0].scatter(mu[::plot_every, 0], mu[::plot_every, 1],
                   c=color[::plot_every], s=s, alpha=alpha)
    if square:
        axs[0].axis('square')
    axs[0].set_title('mu')
    im = axs[1].scatter(logvar[::plot_every, 0], logvar[::plot_every, 1],
                        c=color[::plot_every], s=s, alpha=alpha)
    axs[1].set_title('logvar')
    if square:
        axs[1].axis('square')
    fig.suptitle(name)
    fig.subplots_adjust(right=0.85)
    cbar_ax = fig.add_axes([0.90, 0.15, 0.04, 0.7])
    fig.colorbar(im, cax=cbar_ax)


def plot_embedding_classes(mu, logvar, dict_type_index, prms, dl,
                           plot_every=10, separate_ax=False, square=False,
                           s=0.1, alpha=1, dim=2, ax_logvar=0, figsize=(9, 6)):
    types = dict_type_index.keys()

    if dim == 2:
        if separate_ax:
            mu_min, mu_max = np.min(mu, axis=0), np.max(mu, axis=0)
            logvar_min = np.min(logvar, axis=0)
            logvar_max = np.max(logvar, axis=0)
            plt.figure(figsize=figsize)
            for i, type_ in enumerate(types):
                plt.subplot(2, len(types), i+1)
                plt.scatter(mu[dict_type_index[type_], 0][::plot_every],
                            mu[dict_type_index[type_], 1][::plot_every],
                            s=s, alpha=alpha)
                plt.axis((mu_min[0], mu_max[0], mu_min[1], mu_max[1]))
                plt.title(f'mu {type_}')
            for i, type_ in enumerate(types):
                plt.subplot(2, len(types), i+1+len(types))
                plt.scatter(logvar[dict_type_index[type_], 0][::plot_every],
                            logvar[dict_type_index[type_], 1][::plot_every],
                            s=s, alpha=alpha)
                plt.axis((logvar_min[0], logvar_max[0],
                          logvar_min[1], logvar_max[1]))
                plt.title(f'logvar {type_}')
            plt.tight_layout()
        else:
            plt.figure(figsize=figsize)
            plt.subplot(121)
            for type_ in types:
                plt.scatter(mu[dict_type_index[type_], 0][::plot_every],
                            mu[dict_type_index[type_], 1][::plot_every],
                            label=type_, s=s, alpha=alpha)
            plt.legend(markerscale=1/s)
            if square:
                plt.axis('square')
            plt.title('Repartition of different types (mu)')
            plt.subplot(122)
            for type_ in types:
                plt.scatter(logvar[dict_type_index[type_], 0][::plot_every],
                            logvar[dict_type_index[type_], 1][::plot_every],
                            label=type_, s=s, alpha=alpha)
            plt.legend(markerscale=1/s)
            plt.title('Repartition of different types (logvars)')
            if square:
                plt.axis('square')
            plt.tight_layout()

    if dim == 3:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection='3d')
        for type_ in types:
            ax.scatter(mu[dict_type_index[type_], 0][::plot_every],
                       mu[dict_type_index[type_], 1][::plot_every],
                       logvar[dict_type_index[type_],
                              ax_logvar][::plot_every],
                       label=type_, s=s, alpha=alpha)
        ax.set_xlabel('mu_x')
        ax.set_ylabel('mu_y')
        if ax_logvar == 0:
            ax.set_zlabel('logvar_x')
        else:
            ax.set_zlabel('logvar_y')
        ax.legend(markerscale=1/s)
        ax.set_title('Repartition of different types')


def visualize_types(rw_types, names, nb_pr_row=2, nb_row=2, scale=2,
                    figsize=(16, 7)):
    N = nb_pr_row * nb_row
    ncols = 2
    nrows = len(rw_types) // 2 + len(rw_types) % 2
    print('Generating random walks...')
    dict_type_rws = dict(zip(names,
                             [create_batch_rw(n=N, ps=None, nb_process=None,
                                              types=[type_], pbar=False)[0]
                              for type_ in rw_types]))
    fig, axs = plt.subplots(figsize=figsize, nrows=nrows, ncols=ncols)
    axs = np.atleast_2d(axs)
    for i_type, type_name in zip(list(range(len(rw_types))), names):
        rws = dict_type_rws[type_name]
        for i in range(N):
            rw = rws.loc[rws.n == i, ['x', 'y']].values
            pos = np.array([i//(nb_row)*scale, i % (nb_row)*scale])
            rw += pos
            axs[i_type//2, i_type % 2].plot(rw[:, 0], rw[:, 1])
        for i in range(nb_row+1):
            s2 = scale / 2
            axs[i_type//2, i_type % 2].plot([-s2, nb_pr_row*scale-s2],
                                            [(i-0.5)*scale, (i-0.5)*scale],
                                            'k--', alpha=0.5, linewidth=1)
        for i in range(nb_pr_row+1):
            axs[i_type//2, i_type % 2].plot([(i-0.5)*scale, (i-0.5)*scale],
                                            [-scale/2, nb_row*scale-scale/2],
                                            'k--', alpha=0.5, linewidth=1)
        axs[i_type//2, i_type % 2].set_title(type_name)
        axs[i_type//2, i_type % 2].axis('scaled')
    if len(rw_types) % 2 == 1:
        fig.delaxes(axs[(len(rw_types)-1)//2, 1])
    plt.tight_layout()


def plot_rw(ax, RW):
    points = np.array([RW.x, RW.y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    norm = plt.Normalize(RW.t.min(), RW.t.max())
    lc = LineCollection(segments, cmap='viridis', norm=norm)
    lc.set_array(RW.t)
    line = ax.add_collection(lc)


def plot_sample_RWs(RWs, ncols=4, nrows=4, figsize=(16, 4), scale=2,
                    scale_to_box=False):
    fig, ax = plt.subplots(figsize=figsize)
    ns = RWs.n.unique()
    permut = np.random.permutation(len(ns))
    N = ncols * nrows
    ns_chosen = ns[permut[:N]]
    for i, n_i in enumerate(ns_chosen):
        rw = RWs.loc[RWs.n == n_i].copy()
        if scale_to_box:
            div = np.max(np.array([-rw.x.min(), rw.x.max(),
                                   rw.y.max(), -rw.y.min()]))
            rw.x /= div
            rw.y /= div
        rw.x += (i % ncols) * scale
        rw.y += (i // ncols) * scale
        plot_rw(ax, rw)
    for i in range(ncols):
        plt.axvline((i + 0.5) * scale, linestyle='--')
    for i in range(nrows):
        plt.axhline((i + 0.5) * scale, linestyle='--')
    plt.axis('scaled')



# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This module is where functions processing (extracting features) from random
walks are defined.
"""

import multiprocessing as mp

import numpy as np
import pandas as pd
import tqdm

from .rw_features import *
from .batch_generation import *


def feature_processing(c_drop={'t_max', 't_min', 'size', 'is_dt_cst', 'dt'},
                       c_lin_scale={'area'},
                       c_sqrt_scale={'step_max', 'perimeter'}, **kwargs):
    """Feature processing function that is applied to a DataFrame of the raw
    features extracted from random walks.

    Parameters
    ----------
    c_drop : set of the columns we want to drop from a DataFrame.
    c_lin_scale : set of the columns we want to divide by the duration of the
        random walk.
    c_sqrt_scale : set of the columns we want to divide by the square root of
        the duration of the random walk.

    Returns
    -------
    Function that can be applied to a DataFrame.
    """
    if 'windows' in kwargs:
        def get_func(df, ws=list(kwargs['windows'])):
            time_cols = [(df[f't_max_{w}'] -
                          df[f't_min_{w}']).astype(np.float64) for w in ws]
            c_drop_all = list(
                map(lambda x: [x + f'_{w}' for w in ws], list(c_drop)))
            c_drop_all = set([x for l in c_drop_all for x in l])
            c_drop_all.add('n')

            c_lin_scale_all = list(
                map(lambda x: [x + f'_{w}' for w in ws], list(c_lin_scale)))
            c_lin_scale_all = set([x for l in c_lin_scale_all for x in l])

            c_sqrt_scale_all = list(
                map(lambda x: [x + f'_{w}' for w in ws], list(c_sqrt_scale)))
            c_sqrt_scale_all = set([x for l in c_sqrt_scale_all for x in l])

            df = df.drop(labels=list(
                set(df.columns).intersection(c_drop_all)), axis=1)
            c_scale = c_lin_scale_all.union(c_sqrt_scale_all)
            df_noscale = df[df.columns.difference(c_scale)]
            df_linscale = [
                df[df.columns.intersection(
                    set([f'{x}_{w}' for x in list(c_lin_scale)]))].add_suffix(
                    '_t') for w in ws]
            df_sqrtscale = [df[df.columns.intersection(
                set([f'{x}_{w}' for x in list(c_sqrt_scale)]))].add_suffix(
                '_sqrt') for w in ws]
            df_scaled = (pd.merge(df_linscale[i].divide(time_cols[i],
                                                        axis='index'),
                                  df_sqrtscale[i].divide(np.sqrt(time_cols[i]),
                                                         axis='index'),
                                  left_index=True, right_index=True)
                         for i in range(len(time_cols)))
            return pd.concat((df_noscale, *df_scaled),
                             axis=1).sort_index(axis=1)
        return get_func
    else:
        def get_func(df):
            c_scale = c_lin_scale.union(c_sqrt_scale)
            time_col = (df['t_max'] - df['t_min']).astype(np.float64)
            df = df.drop(labels=list(
                set(df.columns).intersection(c_drop)), axis=1)
            df_noscale = df[df.columns.difference(c_scale)]
            df_linscale = df[c_lin_scale].add_suffix('_t')
            df_sqrtscale = df[c_sqrt_scale].add_suffix('_sqrt')
            df_scaled = pd.merge(df_linscale.divide(time_col, axis='index'),
                                 df_sqrtscale.divide(np.sqrt(time_col),
                                                     axis='index'),
                                 left_index=True, right_index=True)
            return pd.concat((df_noscale, df_scaled),
                             axis=1).sort_index(axis=1)
        return get_func


def get_features_from_group(args):
    """Function introduced to allow the parametrization of the feature
    extraction : the bool zero_time, if True, makes sure that the starting time
    of the random walk is 0.
    Single parameter to be able to use this function with multiprocessing imap.
    """
    RW_df, zero_time = args
    return get_all_features(RandomWalk(RW_df, zero_time=zero_time))


def extract_features(RWs, nb_process=4, func_feat_process=None, pbar=True):
    """Extracts features from a pandas DataFrame collecting different
    trajectories of random walks.

    Parameters
    ----------
    RWs : pandas DataFrame. Columns : n (Index of the random walk), t, and
        dimensions (x, and/or, y, and/or z).
    nb_process : int or None. Number of processes to use if not None.
    func_feat_process : function to apply to raw features extracted from the
        random walk. Use case : to get rid of unused features in the VAE.

    Returns
    -------
    df : pandas DataFrame of the features extracted from trajectories.
        Index is the id of the trajectory, columns are the names of the
        features extracted.
    """
    df_trajs = RWs.groupby('n')
    n_trajs = df_trajs.agg('count').count().x.astype(int)
    if nb_process is None:
        if pbar:
            mes = 'creating rws'
            list_RWobj = [RandomWalk(group, zero_time=True)
                          for _, group in tqdm.tqdm_notebook(df_trajs,
                                                             total=n_trajs,
                                                             desc=mes)]
            raw_features = list(map(get_all_features, tqdm.tqdm_notebook(
                                list_RWobj, total=n_trajs,
                                desc='extracting features')))
        else:
            list_RWobj = [RandomWalk(group, zero_time=True)
                          for _, group in df_trajs]
            raw_features = list(map(get_all_features, list_RWobj))
    else:
        list_args = [(group.copy(), True) for _, group in df_trajs]
        if pbar:
            with mp.Pool(nb_process) as p:
                raw_features = list(tqdm.tqdm_notebook(
                    p.imap(get_features_from_group, list_args),
                    total=n_trajs,
                    desc='creating objects and extracting features'))
        else:
            with mp.Pool(nb_process) as p:
                raw_features = list(p.imap(get_features_from_group, list_args))
    df = pd.DataFrame.from_dict(raw_features)
    df['n'] = RWs.n.unique()
    df.set_index('n', inplace=True)
    if func_feat_process is not None:
        df = func_feat_process(df)
    return df


def create_and_extract(args):
    """Function that extracts features from a single random walk.
    Used for multiprocessing with a generator.
    """
    rw, rw_dict_prm, kwargs = args
    if 'windows' in kwargs:
        ws = kwargs['windows']
        N = len(rw)
        rw_obj = RandomWalk(rw, zero_time=True, check_useless=False)
        feat_scales = [get_all_features(rw_obj,
                                        id_min=int(N/2-w/2),
                                        id_max=int(N/2+w/2)) for w in ws]
        rw_feat = {}
        for i, w in enumerate(ws):
            rw_feat.update({f'{k}_{w}': v for k, v in feat_scales[i].items()})
    elif 'time' in kwargs:
        rw_feat = get_all_features(RandomWalk(rw, zero_time=True),
                                   time_evol=True,
                                   check_useless=False, **kwargs)
    else:
        rw_feat = get_all_features(RandomWalk(rw, zero_time=True,
                                              check_useless=False))
    rw_feat['n'] = rw.n.iloc[0]
    if not kwargs['get_rw']:
        rw = None
    return rw_feat, rw_dict_prm, rw


def features_creation(n=1000, types=[(RW_gauss_dist,
                                      {'d_l': ('float', 'uni', 0.01, 0.1),
                                       'T_max': ('float', 'exp', 0.1, 1)})],
                      ps=[1], get_rw=False, chunksize=1, nb_pos_min=3,
                      nb_process=None, func_feat_process=None, jump_max=10,
                      **kwargs):
    """Creates and directly extracts features from specified types of random
    walks. Avoids the creation of a pandas DataFrame of the trajectories :
    useful for lowering the RAM usage.

    Parameters
    ----------
    n : int, the number of trajectories we want to generate.
    type : list that describes which random walks we generate.
        Each item is a tuple with 2 elements.
        - The first is the function (see rw_simulation) which characterizes
        which type of random walk it produces.
        - The second is a dictionary whose keys are the parameters we want to
        pass to the random walk function and values the parameters describing
        how those parameter values are generated with `generate_random_number`.
    ps : parameter passed to `np.random.choice` as p, should have same
        dimension as type. Controls the probability distribution of choosing
        some type.
    get_rw : bool, optional. Whether we want to retrieve trajectories or not.
    nb_process : int or None. If int, the number of processes to use to benefit
        from parallelization.
    func_feat_process : function to apply to raw features extracted from the
        random walk. Use case : to get rid of unused features in the VAE.

    Returns
    -------
    df : pandas DataFrame of the features extracted from trajectories.
        Index is the id of the trajectory, columns are the names of the
        features extracted.
    rw_prms : dict that carries info on each trajectory generated. Each value
        is the index in df of the trajectory, each key is a dictionary
        of the parameters used to generate the random walk (except for
        default parameters). It also carries the name of the function which
        generated the random walk.
    df_rws : None or pandas DataFrame depending on get_rw, raw data of the
        trajectories.
    """
    kwargs['get_rw'] = get_rw
    rw_generator = rw_feature_generator(n, types=types, ps=ps,
                                        nb_pos_min=nb_pos_min,
                                        jump_max=jump_max, **kwargs)
    desc = 'creating and extracting rws features'
    if nb_process is None:
        raw_data = list(map(create_and_extract,
                            tqdm.tqdm_notebook(rw_generator,
                                               total=n, desc=desc)))
    else:
        with mp.Pool(nb_process) as p:
            raw_data = list(tqdm.tqdm_notebook(
                p.imap(create_and_extract, rw_generator, chunksize=chunksize),
                total=n, desc=desc))
    raw_features = []
    rw_prms = {}
    list_rws = []
    for i in range(len(raw_data)):
        raw_features.append(raw_data[i][0])
        rw_prms[i] = raw_data[i][1]
        if get_rw:
            list_rws.append(raw_data[i][2])
    df = pd.DataFrame.from_dict(raw_features)
    if func_feat_process is not None:
        df = func_feat_process(df)
    if get_rw:
        df_rws = pd.concat(list_rws).reset_index(drop=True)
    else:
        df_rws = None
    return df, rw_prms, df_rws



# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .abc import *
import numpy as np
from collections.abc import Iterable, Sequence, Set
import logging
import functools

__all__ = ['GenericAttribute', 'Attribute']

__all__.append('WithLogger')
class WithLogger(object):
    __slots__ = ('_logger',)
    def __init__(self):
        self._logger = None
    @property
    def logger(self):
        """
        logging.Logger:
            logger instance of the module the object is defined in,
            unless explicitly set
        """
        if self._logger is None:
            return logging.getLogger(type(self).__module__)
        else:
            return self._logger
    @logger.setter
    def logger(self, logger):
        self._logger = logger

__analyzer_node_init_args__ = ('parent',)

__all__.append('AnalyzerNode')
class AnalyzerNode(WithLogger):
    __slots__ = ('_parent',)
    def __init__(self, parent=None):
        WithLogger.__init__(self)
        self._parent = parent
    @property
    def _eldest_parent(self):
        parent = self
        while True:
            if parent._parent is None:
                break
            else:
                parent = parent._parent
                if not isinstance(parent, AnalyzerNode):
                    #assert isinstance(parent, RWAnalyzer)
                    break
        return parent
    def _bear_child(self, cls, *args, **kwargs):
        assert issubclass(cls, AnalyzerNode)
        kwargs['parent'] = self
        return cls(*args, **kwargs)
    @property
    def initialized(self):
        return True
    @property
    def logger(self):
        if self._logger is None:
            parent = self._eldest_parent
            if parent is None or parent._logger is None:
                return WithLogger.logger.fget(self)
            else:
                return parent.logger
        else:
            return WithLogger.logger.fget(self)
    @logger.setter
    def logger(self, logger):
        WithLogger.logger.fset(self, logger)


__all__.append('Initializer')
class Initializer(AnalyzerNode):
    __slots__ = ('_specialize',)
    def __init__(self, attribute_setter, parent=None):
        AnalyzerNode.__init__(self, parent)
        self._specialize = attribute_setter
    def specialize(self, attribute_cls, *cls_args, **cls_kwargs):
        cls_kwargs['parent'] = self._parent
        self._specialize(attribute_cls(*cls_args, **cls_kwargs))
    @property
    def initialized(self):
        return False
    @property
    def reified(self):
        return False

GenericAttribute.register(Initializer)


__all__.append('selfinitializing_property')
def selfinitializing_property(attr_name, getter, setter, metacls=None, doc=None):
    def initializer(self, cls):
        if cls is None:
            setter(self, None)
        elif callable(cls):
            if isinstance(cls, InitializerMethod):
                cls.assign( self )
                return
            if metacls:
                def typechecked_setter(obj):
                    if not isinstance(obj, metacls):
                        raise TypeError("the argument to attribute '{}' is not a {}".format(attr_name, metacls.__name__))
                    setter(self, obj)
                setter(self, cls( typechecked_setter, parent=self ))
            else:
                setter(self, cls( setter, parent=self ) )
            if not isinstance(getter(self), Initializer):
                raise TypeError("the argument to attribute '{}' is not an Initializer".format(attr_name))
        else:
            raise TypeError("the argument to attribute '{}' is not an Initializer".format(attr_name))
    return property(getter, initializer, doc=doc)


__all__.append('null_index')
def null_index(i):
    if i is None:
        return True
    elif isinstance(i, Iterable):
        # expect a single-element sequence which value should be 0,
        # idealy of integer type, but no type check is performed here
        # because of the numerous integer types in NumPy
        it = iter(i)
        try:
            i = next(it)
        except StopIteration:
            # empty
            return False
        else:
            if i != 0:
                return False
            try:
                i = next(it)
            except StopIteration:
                return True
            else:
                # multiple indices; return False even if all are null
                return False
    elif callable(i):
        return i(0)
    else:
        return i == 0


__all__.append('indexer')
def indexer(i, it, return_index=False, has_keys=False):
    if i is None:
        if return_index:
            if has_keys:
                for k in it:
                    yield k, it[k]
            else:
                yield from enumerate(it)
        else:
            if has_keys:
                for k in it:
                    yield it[k]
            else:
                yield from it
    elif callable(i):
        if has_keys:
            if return_index:
                for j in it:
                    if i(j):
                        yield j, item
            else:
                for j in it:
                    if i(j):
                        yield item
        else:
            if return_index:
                for j, item in enumerate(it):
                    if i(j):
                        yield j, item
            else:
                for j, item in enumerate(it):
                    if i(j):
                        yield item
    elif isinstance(i, (bytes, str)):
        if not has_keys:
            raise TypeError('string index for indexing a collection with no keys')
        if return_index:
            yield i, it[i]
        else:
            yield it[i]
    elif isinstance(i, (Sequence, np.ndarray)):
        if has_keys:
            if return_index:
                for k in i:
                    yield k, it[k]
            else:
                for k in i:
                    yield it[k]
        else:
            # no need for `__getitem__`, but iteration follows the ordering in `i`
            postponed = dict()
            j, it = -1, iter(it)
            for k in i:
                try:
                    while j<k:
                        if j in i:
                            postponed[j] = item
                        j, item = j+1, next(it)
                except StopIteration:
                    raise IndexError('index is out of bounds: {}'.format(k))
                if j != k:
                    try:
                        item = postponed.pop(k)
                    except KeyError:
                        if k < 0:
                            raise IndexError('negative values are not supported in a sequence of indices')
                        else:
                            raise IndexError('duplicate index: {}'.format(k))
                if return_index:
                    yield k, item
                else:
                    yield item
    elif isinstance(i, Set):
        i = set(i) # copy and make mutable
        if has_keys:
            for j in it:
                if j in i:
                    if return_index:
                        yield j, it[j]
                    else:
                        yield it[j]
                    i.remove(j)
        else:
            for j, item in enumerate(it):
                if j in i:
                    if return_index:
                        yield j, item
                    else:
                        yield item
                    i.remove(j)
        if i:
            raise IndexError(('some indices are out of bounds: '+', '.join(['{}'])).format(*tuple(i)))
    elif np.isscalar(i):
        if has_keys:
            try:
                item = it[i]
            except KeyError:
                raise IndexError('index is out of bounds: {}'.format(i)) from None
        else:
            it = iter(it)
            if i == -1:
                try:
                    while True:
                        item = next(it)
                except StopIteration:
                    pass
            else:
                j = -1
                try:
                    while j<i:
                        j, item = j+1, next(it)
                except StopIteration:
                    raise IndexError('index is out of bounds: {}'.format(i))
        if return_index:
            yield i, item
        else:
            yield item
    else:
        raise TypeError('unsupported index type')


__all__.append('first')
def first(it, **kwargs):
    if callable(it):
        it = it(**kwargs)
    return next(iter(it))


__all__.append('single')
def single(it, **kwargs):
    if callable(it):
        it = it(**kwargs)
    it = iter(it)
    elem = next(it)
    try:
        next(it)
    except StopIteration:
        # good
        pass
    else:
        raise RuntimeError('not a singleton') from None
    return elem


__all__.append('Proxy')
class Proxy(object):
    __slots__ = ('_proxied',)
    def __init__(self, proxied):
        self._proxied = proxied
    def __len__(self):
        return self._proxied.__len__()
    def __iter__(self):
        return self._proxied.__iter__()
    @property
    def _parent(self):
        return self._proxied._parent
    @_parent.setter
    def _parent(self, par):
        self._proxied._parent = par
    def __getattr__(self, attrname):
        if attrname == '_proxied':
            # prevent copy.copy from entering infinite recursion
            return
        return getattr(self._proxied, attrname)
    def __setattr__(self, attrname, val):
        if attrname == '_proxied':
            object.__setattr__(self, '_proxied', val)
        else:
            setattr(self._proxied, attrname, val)


__all__.append('InitializerMethod')
class InitializerMethod(object):
    """
    Useful for explicit typing.
    """
    __slots__ = ('attrname', 'method', 'args', 'kwargs')
    def __init__(self, method, attrname=None):
        self.method = method
        if attrname is None:
            attrname = method.__module__.split('.')[-1]
        self.attrname = attrname
        self.args = self.kwargs = None
    def __call__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self
    def assign(self, parent_analyzer):
        if not (isinstance(self.args, tuple) and isinstance(self.kwargs, dict)):
            raise RuntimeError('initializer method has not been called')
        parent_attribute = getattr(parent_analyzer, self.attrname)
        self.method(parent_attribute, *self.args, **self.kwargs)
    def reassign(self, parent_analyzer):
        import importlib
        mod = importlib.import_module(self.method.__module__)
        try:
            clsname = str(self.method)[::-1].split('.', 1)[-1][::-1].split()[-1]
            cls = getattr(mod, clsname)
            assert isinstance(cls, type) and issubclass(cls, Initializer)
        except:
            raise AttributeError("cannot find a proper initializer for the '{}' attribute".format(self.attrname))
        attr = getattr(type(parent_analyzer), self.attrname)
        # TODO: add a freset method to the properties returned by selfinitializing_property
        attr.fset(parent_analyzer, cls)
        self.assign( parent_analyzer )

__all__.append('initializer_method')
def initializer_method(method, attrname=None):
    """
    Properly set the docstring for the wrapped method.
    """
    meth = InitializerMethod(method, attrname)
    @functools.wraps(method)
    def wrapped_method(*args, **kwargs):
        return meth(*args, **kwargs)
    return wrapped_method





# -*- coding: utf-8 -*-

# Copyright © 2018, Alexander Serov

import logging

import numpy as np
from numpy import exp, log
from scipy import integrate
from scipy.special import gamma, gammainc, gammaln

# Constants

# prec = 1e-32
# max_terms = int(1e5)


def calculate_marginalized_integral(zeta_t, zeta_sp, p, v, E, rel_loc_error, zeta_a=[0, 0], factor_za=0, lamb='int'):
    """
    Calculate the marginalized lambda integral
    >>>
    Integrate[gamma_inc[p, arg * rel_loc_error] * arg ** (-p), {lambda, 0, 1}]
    >>>
    for the given values of zeta_t and zeta_sp.

    Here:
    arg = (v + factor_za * zeta_a**2 + E * (zeta_t - zeta_a - lambda * zeta_sp)**2);
    IMPORTANT: gamma_inc --- is the normalized lower incomplete gamma function;
    rel_loc_error = n * V / (4 * sigma_L^2) --- inverse relative localization error,


    Input:
    zeta_t and zeta_sp are vectors of length dim (x, y, z).
    All other parameters are scalars.
    abs_tol --- absolute tolerance for integral calculations.
    lamb = float in [0, 1] or 'int'. If a float is given, the integrated function is just evaluated at the given value of lambda, without integrating.
    """
    atol = 1e-16

    zeta_t, zeta_sp, zeta_a = map(np.asarray, [zeta_t, zeta_sp, zeta_a])
    if zeta_t.ndim > 1 or zeta_sp.ndim > 1:
        raise ValueError("zeta_t and zeta_sp should be 1D vectors")

    zeta_a_norm_sq = np.sum(zeta_a**2)

    def arg(l):
        """lambda-dependent argument of the gamma function and the second term."""
        diff = zeta_t - zeta_a - l * zeta_sp
        # the @ operator raises a syntax error in Py2
        return v + factor_za * zeta_a_norm_sq + E * np.matmul(diff, diff.T)

    def get_integrate_me():
        """Function to integrate with and without localization error."""
        def no_error(l):
            return arg(l) ** (-p)

        def with_error(l):
            return gammainc(p, arg(l) * rel_loc_error) * arg(l) ** (-p)

        if rel_loc_error == np.inf:
            return no_error
        else:
            return with_error

    integrate_me = get_integrate_me()

    # Skip integration if zeta_sp is close to 0
    if np.isclose(np.linalg.norm(zeta_sp), 0, atol=atol):
        return integrate_me(0)
    # Skip integration if a fixed-lambda convention is used
    if lamb != 'int':
        return integrate_me(lamb)

    # Calculate break points
    with np.errstate(invalid='ignore', divide='ignore'):
        lambda_breaks = [lb for lb in np.divide(
            zeta_t - zeta_a, zeta_sp) if lb >= 0.0 and lb <= 1.0]

    # Perform integration
    result = integrate.quad(integrate_me, 0.0, 1.0, points=lambda_breaks,
                            full_output=False, epsabs=atol)
    return result[0]


def calculate_any_lambda_integral(func, break_points=[]):
    """
    DEPRECATED: calculate ratios instead.
    This function allows to calculate more complicated lambda integrals, such as 1D posteriors in 2D.

    Input:
    func(lambda) - function of lambda to integrate,
    break_points - special integration points (i.e. zeta_t/zeta_sp), optional.
    """
    #
    # zeta_t, zeta_sp, zeta_a = map(np.asarray, [zeta_t, zeta_sp, zeta_a])
    # if zeta_t.ndim > 1 or zeta_sp.ndim > 1:
    #     raise ValueError("zeta_t and zeta_sp should be 1D vectors")
    #
    # zeta_a_norm_sq = np.sum(zeta_a**2)

    # def arg(l):
    #     """lambda-dependent argument of the gamma function and the second term."""
    #     diff = zeta_t - zeta_a - l * zeta_sp
    #     # the @ operator raises a syntax error in Py2
    #     return v + factor_za * zeta_a_norm_sq + E * np.matmul(diff, diff.T)
    #
    # def get_integrate_me():
    #     """Function to integrate with and without localization error."""
    #     def no_error(l):
    #         return arg(l) ** (-p)
    #
    #     def with_error(l):
    #         return gammainc(p, arg(l) * rel_loc_error) * arg(l) ** (-p)
    #
    #     if rel_loc_error == np.inf:
    #         return no_error
    #     else:
    #         return with_error
    #
    # integrate_me = get_integrate_me()

    # # Skip integration if zeta_sp is close to 0
    # if np.isclose(np.linalg.norm(zeta_sp), 0, atol=atol):
    #     return integrate_me(0)
    # # Skip integration if a fixed-lambda convention is used
    # if lamb is not 'int':
    #     return integrate_me(lamb)

    # Verify break points
    with np.errstate(invalid='ignore', divide='ignore'):
        lambda_breaks = [lb for lb in break_points if lb >= 0.0 and lb <= 1.0]

    # Perform integration
    result = integrate.quad(func, 0.0, 1.0, points=lambda_breaks,
                            full_output=False, epsabs=atol)
    return result[0]


def calculate_integral_ratio(arg_func_up, arg_func_down, pow_up, pow_down, v, rel_loc_error, break_points=[], lamb='marg', rtol=1e-6, atol=1e-32):
    """
    Calculate the ratio of two similar lambda integrals, each of which has the form
    \int_0^1 d\lambda arg_func(\lambda)**(-pow) * gammainc(pow, rel_loc_error * arg_func(\lambda))

    Input:
    v --- zeta-independent term inside the argument functions. Must be the same upstairs and downstairs. In Bayes factor calculations, this is v := (1 + n_pi * V_pi / n / V).

    Return:
    Natural logarithm of the integral ratio

    Details:
    Calculations are performed differently for the case with and without localization error. This avoids overflow errors. In these cases, both the integrated equations and prefactors are different.
    The integrals are taken over lambda if lamb == 'marg', otherwise they are evaluated at the given lambda.
    """

    x0u = arg_func_up(lamb) if lamb != 'marg' else arg_func_up(0.5)
    x0d = arg_func_down(lamb) if lamb != 'marg' else arg_func_down(0.5)

    if np.any(np.array([pow_up, pow_down]) <= 0) or np.any(np.isnan([pow_up, pow_down])):
        logging.error(
            'Incorrect powers supplied for integral ratio calculations: {}'.format([pow_up, pow_down]))
        return np.nan

    # %% loc_error > 0
    def get_f_with_loc_error(arg_func, pow, x0):
        """Prepare the function for lambda integration.
        Switch summation algorithm based on the argument value at lambda = lamb or 1/2 for marginalized.

        Return:
        f --- a function to integrate,
        ln_prefactor --- a scaling prefactor
        """

        def term(i, l):
            q = arg_func(l)
            res = gammaln(pow) - gammaln(pow + 1 + i) + i * \
                log(q * rel_loc_error) - q * rel_loc_error
            return exp(res)

        q_eval = arg_func(lamb) if lamb != 'marg' else arg_func(0.5)
        q_eval *= rel_loc_error
        if q_eval < pow + 1:
            # print('Low x regime', pow, q_eval)

            def f(l):
                return sum_series(term, rtol, l)
            ln_prefactor = pow * log(rel_loc_error)
            return f, ln_prefactor
        else:
            def f(l):
                q = arg_func(l)
                res = -pow * (log(q) - log(x0)) + log(gammainc(pow, q * rel_loc_error))
                return exp(res)
            ln_prefactor = gammaln(pow) - pow * log(x0)
            return f, ln_prefactor

    def ln_integral_with_loc_error(arg_func, pow, x0):
        """Calculate the following expression for a marginalized or fixed lambda:
        v**k / Gamma[k] * \int_0^1 d l q(l)**(-k) gammainc(k, q(l) * rel_loc_error),
        with q(l) = arg_func(l)
        """
        f, ln_prefactor = get_f_with_loc_error(arg_func, pow, x0)
        if lamb == 'marg':
            ln_res = log(integrate.quad(f, 0, 1, points=break_points, epsrel=rtol)[0])
        else:
            ln_res = log(f(lamb))
        return ln_res + ln_prefactor

    # %% loc_error == 0
    def ln_integral_no_error(arg_func, pow, x0):
        ln_prefactor = gammaln(pow) - pow * log(x0)

        def ln_f(l):
            return -pow * log(arg_func(l) / x0)

        def f(l):
            return exp(ln_f(l))
        if lamb == 'marg':
            ln_res = integrate.quad(f, 0, 1, points=break_points, epsrel=rtol)[0]
            ln_res = log(ln_res)
        else:
            ln_res = ln_f(lamb)
        return ln_res + ln_prefactor

    # %% Compile the ratio
    if ~np.isinf(rel_loc_error):
        log_res = (ln_integral_with_loc_error(arg_func_up, pow_up, x0u)
                   - ln_integral_with_loc_error(arg_func_down, pow_down, x0d))
    else:
        log_res = (ln_integral_no_error(arg_func_up, pow_up, x0u) -
                   ln_integral_no_error(arg_func_down, pow_down, x0d))
    return log_res


def sum_series(term_func, rtol, *args):
    """Sum any series with the first argument of the term_func being the term number.
    Input: function, return - float
    """
    max_terms = int(1e5)
    sum = term = term_func(0, *args)
    for i in range(1, max_terms + 1):
        term = term_func(i, *args)
        sum += term
        if abs(term) <= rtol * abs(sum):
            return sum
    logging.warning(
        'Gamma function series did not converge to the required precision. Achieved precision: {}'.format(abs(term / sum)))
    return sum



# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd
from tramway.core.xyt import crop
import warnings


def random_walk(diffusivity=None, force=None, friction=None, drift=None,
        trajectory_mean_count=100, trajectory_count_sd=0, turnover=None,
        initial_trajectory_count=None, new_trajectory_count=None,
        lifetime_tau=None, lifetime=None, single=False,
        box=(0., 0., 1., 1.), duration=10., time_step=.05, minor_step_count=99,
        reflect=False, full=False, count_outside_trajectories=None,
        viscosity=None):
    r"""
    Generate random walks.

    Consider also the alternative generator
    :func:`~tramway.helper.simulation.categoricaltrap.random_walk_2d`.

    The `trajectory_mean_count` and `trajectory_count_sd` arguments are not
    compatible with the `initial_trajectory_count` and `new_trajectory_count`
    arguments.
    Use either pair of arguments.

    *new in 0.6*: argument `friction` replaces `viscosity`, and a warning is
    issued if `viscosity` is used.

    The units for `force` and `friction` can be adapted as long as the ratio
    of force over friction is expressed as :math:`\mu m s^{-1}` (and drift is
    not defined).

    Arguments:

        diffusivity (callable or float): if callable, takes a coordinate vector
            (:class:`~numpy.ndarray`) and time (float), and returns the local
            diffusivity (float) with :math:`\mu m^2.s^{-1}` as a unit

        force (callable): takes a coordinate vector (:class:`~numpy.ndarray`)
            and time (float) and returns the local force vector
            (:class:`~numpy.ndarray`) with :math:`k_{\rm{B}}T\mu m^{-1}` as a
            unit

        friction (callable or float): if callable, takes a coordinate vector
            (:class:`~numpy.ndarray`) and time (float), and returns the local
            friction (float) divided by :math:`k_{\rm{B}}T`,
            equal to the inverse of the diffusivity under equilibrium

        drift (callable): takes a coordinate vector (:class:`~numpy.ndarray`)
            and time (float) if `force` is not defined,
            or the same two arguments plus the diffusivity (float) and force
            (:class:`~numpy.ndarray`) otherwise, and returns the local drift;
            this lets `friction` unevaluated

        trajectory_mean_count (int or float): average number of active
            trajectories at any time

        trajectory_count_sd (int or float): standard deviation of the number of
            active trajectories

        turnover (float): fraction of trajectories that will end at each time
            step; **deprecated**

        initial_trajectory_count (int): initial number of active trajectories

        new_trajectory_count (int or callable): number of newly generated
            trajectories; if `new_trajectory_count` is ``callable``, then it
            takes the time of a step as input and returns the count for this
            step as an ``int``

        lifetime_tau (float): trajectory lifetime constant in seconds

        lifetime (callable or float): trajectory lifetime in seconds;
            if `lifetime` is a ``float``, then it acts like `lifetime_tau`;
            if `lifetime` is ``callable``, then it takes as input the initial
            time of the trajectory

        single (bool): allow single-point trajectories

        box (array-like): origin and size of the space bounding box

        duration (float): duration of the simulation in seconds

        time_step (float): duration between two consecutive observations

        minor_step_count (int): number of intermediate unobserved steps

        reflect (bool): reflect the minor steps that would otherwise leave the
            bounding box

        full (bool): include locations that are outside the bounding box and
            times that are posterior to `duration`

        count_outside_trajectories (bool): include trajectories that have left
            the bounding box in determining the number of trajectories at each
            observation step; **deprecated**

    Returns:

        pandas.DataFrame: simulated trajectories with 'n' the trajectory index,
            't' the time and with other columns for location coordinates

    """
    if turnover is not None:
        warnings.warn('`turnover` is deprecated', DeprecationWarning)
    if count_outside_trajectories is False:
        warnings.warn('`count_outside_trajectories` is deprecated', DeprecationWarning)
    _box = np.asarray(box)
    dim = int(_box.size / 2)
    support_lower_bound = _box[:dim]
    support_size = _box[dim:]
    support_upper_bound = support_lower_bound + support_size
    # default diffusivity and drift maps
    #def null_scalar_map(xy, t):
    #    return 0.
    #def null_vector_map(xy, t):
    #    return np.zeros((dim,))
    if callable(diffusivity):
        pass
    elif np.isscalar(diffusivity) and 0 < diffusivity:
        _D = diffusivity
        diffusivity = lambda x, t: _D
    else:
        raise ValueError('`diffusivity` must be callable or a positive float')
    if viscosity is not None:
        if friction is None:
            friction = viscosity
            warnings.warn(r"argument name 'viscosity' is misleading as it actually represents the friction divided by kT")
        else:
            warnings.warn(r"argument 'viscosity' is ignored")
    if force and not callable(force):
        raise TypeError('`force` must be callable')
    if drift:
        _drift = drift
        if force:
            drift = lambda x, t, D: _drift(x, t, D, force(x, t))
        else:
            drift = lambda x, t, D: _drift(x, t)
    elif force:
        if friction is None:
            drift = lambda x, t, D: D * force(x, t)
        elif callable(friction):
            drift = lambda x, t, D: force(x, t) / friction(x, t)
        elif np.isscalar(friction) and 0 < friction:
            drift = lambda x, t, D: force(x, t) / friction
        else:
            raise ValueError('`friction` must be callable or a positive float')
    else:
        drift = lambda x, t, D: 0
    #
    N = int(round(float(duration) / time_step)) # number of observed steps
    min_step_count = 1 if single else 2 # minimum number of observed steps per trajectory
    if callable(lifetime):
        lifetime_tau = None
    elif lifetime:
        lifetime_tau, lifetime = lifetime, None
    if lifetime_tau:
        trajectory_count_sd = None
    elif trajectory_count_sd:
        lifetime_tau = float(trajectory_count_sd) * time_step
    else:
        lifetime_tau = 4. * time_step
    # check
    if lifetime:
        assert callable(lifetime)
    else:
        assert bool(lifetime_tau)
    # number of trajectories at each time step
    K = None
    if new_trajectory_count is None:
        if trajectory_count_sd:
            K = np.rint(np.random.randn(N) * trajectory_count_sd + trajectory_mean_count).astype(int)
        else:
            K = np.full(N, trajectory_mean_count, dtype=int)
    elif initial_trajectory_count is None:
        warnings.warn('in combination with `new_trajectory_count`, use `initial_trajectory_count` instead of `trajectory_mean_count`')
        initial_trajectory_count = trajectory_mean_count
    k = np.zeros(N, dtype=int)
    # starting time and duration of the trajectories
    time_support = []
    t = 0.
    for i in range(N):
        t += time_step
        if K is None:
            if i == 0:
                k_new = initial_trajectory_count
            elif callable(new_trajectory_count):
                k_new = new_trajectory_count(t)
            else:
                k_new = new_trajectory_count
        else:
            k_new = K[i] - k[i]
        if k_new <= 0:
            if k_new != 0:
                print('{:d} excess trajectories at step {:d}'.format(-k_new, i))
            continue
        if lifetime:
            _lifetime = np.array([ lifetime(t) for j in range(k_new) ])
        else:
            _lifetime = -np.log(1 - np.random.rand(k_new)) * lifetime_tau
        _lifetime = np.rint(_lifetime / time_step).astype(int) + 1
        _lifetime = _lifetime[min_step_count <= _lifetime]
        time_support.append((t, _lifetime))
        if _lifetime.size == 0:
            continue
        _lifetimes, _count = np.unique(_lifetime, return_counts=True)
        _lifetime = np.zeros(_lifetimes.max(), dtype=_count.dtype)
        _lifetime[_lifetimes-1] = _count
        if K is not None:
            _count = np.flipud(np.cumsum(np.flipud(_lifetime)))
            k[i:min(i+_count.size,k.size)] += _count[:min(k.size-i,_count.size)]
    #
    actual_time_step = time_step / float(minor_step_count + 1)
    total_location_count = sum([ np.sum(_lifetime) for _, _lifetime in time_support ])
    # generate the trajectories
    N = np.empty((total_location_count, ), dtype=int)       # trajectory index
    X = np.empty((total_location_count, dim), dtype=_box.dtype) # spatial coordinates
    T = np.empty((total_location_count, ), dtype=float)     # time
    i, n = 0, 0
    for t0, lifetimes in time_support:
        k = lifetimes.size # number of new trajectories at time t
        X0 = np.random.rand(k, dim) * support_size + support_lower_bound # initial coordinates
        for x0, _lifetime in zip(X0, lifetimes): # for each new trajectory
            n += 1
            N[i] = n
            X[i] = x = x0
            T[i] = t = t0
            i += 1
            # from here the main code (``not reflect``) is duplicated to reduce the number of if-tests
            if reflect:
                # duplicated code
                for j in range(1,_lifetime):
                    for _ in range(minor_step_count+1):
                        D = diffusivity(x, t)
                        A = drift(x, t, D)
                        dx = actual_time_step * A + \
                            np.sqrt(actual_time_step * 2. * D) * np.random.randn(dim)
                        x = x + dx
                        # additional code
                        above = support_upper_bound < x
                        x[above] = 2*support_upper_bound[above] - x[above]
                        below = x < support_lower_bound
                        if np.any(below & above):
                            raise NotImplementedError('the jump is so large that its reflection also exceeds the opposite bound')
                        x[below] = 2*support_lower_bound[below] - x[below]
                        #
                        t = t + actual_time_step
                    t = t0 + j * time_step # moderate numerical precision errors
                    N[i] = n
                    X[i] = x
                    T[i] = t
                    i += 1
            else:
                # reference code
                for j in range(1,_lifetime):
                    for _ in range(minor_step_count+1):
                        D = diffusivity(x, t)
                        A = drift(x, t, D)
                        dx = actual_time_step * A + \
                            np.sqrt(actual_time_step * 2. * D) * np.random.randn(dim)
                        x = x + dx
                        t = t + actual_time_step
                    t = t0 + j * time_step # moderate numerical precision errors
                    N[i] = n
                    X[i] = x
                    T[i] = t
                    i += 1
    # format the data as a dataframe
    columns = 'xyz'
    if dim <= 3:
        xcols = [ d for d in columns[:dim] ]
    else:
        xcols = [ 'x'+str(i) for i in range(dim) ]
    points = pd.DataFrame(N, columns=['n']).join(
        pd.DataFrame(X, columns=xcols)).join(
        pd.DataFrame(T, columns=['t']))
    # post-process
    if not full and not reflect:
        points = crop(points, _box)
        points = points.loc[points['t'] <= duration+time_step*.1]
        if not single:
            n, count = np.unique(points['n'].values, return_counts=True)
            n = n[count == 1]
            if n.size:
                points = points.loc[~points['n'].isin(n)]
        points.index = np.arange(points.shape[0])
    return points


def add_noise(points, sigma, copy=False):
    columns = [ c for c in points.columns if c not in ('n', 't') ]
    dim = len(columns)
    npoints = points.shape[0]
    if copy:
        points = points.copy()
    points[columns] += sigma * np.random.randn(npoints, dim)
    return points


def truth(cells, t=None, diffusivity=None, force=None, potential=None, **kwargs):
    """
    Generate maps for the true diffusivity/force distribution.

    If `cells` is a :class:`~tramway.tessellation.base.Partition` object,
    :func:`~tramway.inference.base.distributed` is called on `cells` and with the trailing
    keyword arguments to build a :class:`~tramway.inference.base.FiniteElements` object.

    Arguments:

        cells (Partition or FiniteElements): distributed data ready for inference

        t (float): time as understood by `diffusivity` and `force`

        diffusivity (callable): admits the coordinates of a single location (array-like)
            and time (float, if `t` is defined) and returns the local diffusivity (float)

        force (callable): admits the coordinates of a single location (array-like)
            and time (float, if `t` is defined) and returns the local force (array-like)

        potential (callable): admits the coordinates of a single location (array-like)
            and time (float, if `t` is defined) and returns the local potential energy (float)

    Returns:

        pandas.DataFrame: diffusivity/force maps
    """
    try:
        cells.cells
    except AttributeError:
        import tramway.inference as ti
        cells = ti.distributed(cells, **kwargs)
    I, DVF = [], []
    for i in cells.cells:
        cell = cells.cells[i]
        if not I:
            dim = cell.center.size
        I.append(i)
        if diffusivity is None:
            D = []
        elif t is None:
            D = [diffusivity(cell.center)]
        else:
            D = [diffusivity(cell.center, t)]
        if force is None:
            F = []
        elif t is None:
            F = force(cell.center)
        else:
            F = force(cell.center, t)
        if potential is None:
            V = []
        elif t is None:
            V = [potential(cell.center)]
        else:
            V = [potential(cell.center, t)]
        DVF.append(np.concatenate((D, V, F)))
    DVF = np.vstack(DVF)
    if diffusivity is None:
        columns = []
    else:
        columns = [ 'diffusivity' ]
    if potential is not None:
        columns.append('potential')
    if force is not None:
        columns += [ 'force x' + str(col+1) for col in range(dim) ]
    return pd.DataFrame(index=I, data=DVF, columns = columns)





# -*- coding: utf-8 -*-

# Copyright © 2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import AnalyzerNode, single
from ..artefact import Analysis
import numpy as np
import pandas as pd
import plotly # missing dependencies: plotly nbformat
import plotly.graph_objects as go

"""
Plotly interfaces currently are highly experimental
and may suffer from numerous bugs.

The `plotly` dependencies are not enforced and an
:class:`ImportError` is raised if the library cannot
be found.
"""

class Plotly(AnalyzerNode):
    """
    Plotly interface for maps.

    It does not access other attributes of the :class:`RWAnalyzer`,
    and thus can be safely used from any :class:`RWAnalyzer` object:

        from tramway.analyzer import RWAnalyzer

        RWAnalyzer().mapper.plotly.plot_surface(...)

    or:

        from tramway.analyzer.mapper.plotly import Plotly

        Plotly.plot_surface(...)

    """

    def plot_surface(self, maps, feature, sampling=None, fig=None,
            row=None, col=None, colormap='viridis', title=None,
            colorbar=None, resolution=200, **kwargs):
        """
        Plot a 2D map as a colored 3D surface.
        """
        surface_kwargs = kwargs
        surface_kwargs['colorscale'] = kwargs.pop('colorscale', colormap)

        figure_kwargs = {}
        for kw in ('scene',):
            try:
                arg = kwargs.pop(kw)
            except KeyError:
                pass
            else:
                figure_kwargs[kw] = arg

        if fig is not None:
            if not (row is None and col is None):
                # subplots
                figure_kwargs.update(dict(
                    row=1 if row is None else row+1,
                    col=1 if col is None else col+1,
                    ))

        if colorbar is not None:
            surface_kwargs['colorbar'] = colorbar

        if isinstance(maps, Analysis):
            if sampling is None:
                sampling = maps.get_parent()
            maps = maps.data
        if isinstance(sampling, Analysis):
            sampling = sampling.data

        tessellation = sampling.tessellation
        try:
            tessellation = tessellation.spatial_mesh
        except AttributeError:
            pass

        xlim, ylim = sampling.bounding_box[['x', 'y']].values.T
        step = max(xlim[1] - xlim[0], ylim[1] - ylim[0]) / resolution
        x = np.arange(xlim[0], xlim[1], step)
        y = np.arange(ylim[0], ylim[1], step)
        x, y = np.meshgrid(x, y)

        pts = np.stack((x.ravel(), y.ravel()), axis=1)
        partition_kwargs = dict(sampling.param.get('partition', {}))
        for kw in list(partition_kwargs.keys()):
            if kw.startswith('time_'):
                del partition_kwargs[kw]
            elif kw in ('min_location_count', 'knn'):
                del partition_kwargs[kw]
        cell_ix = tessellation.cell_index(
                pd.DataFrame(pts, columns=['x', 'y']),
                **partition_kwargs)

        map_ = maps[feature]
        if 1<map_.shape[1]:
            map_ = map_.pow(2).sum(1).apply(np.sqrt)
        else:
            map_ = map_[feature]

        z = np.zeros(len(pts), dtype=map_.dtype)

        if isinstance(cell_ix, tuple):
            pts_, cells = cell_ix
            assert np.all(np.unique(pts_) == np.arange(len(pts)))
            # brute force accum array
            n = np.zeros(z.shape, dtype=int)
            for i in map_.index:
                k = pts_[cells==i]
                if k.size:
                    z[k] += map_[i]
                    n[k] += 1
            ok = 0<n
            z[1<n] /= n[1<n]
        else:
            ok = 0<=cell_ix
            map__ = np.full(tessellation.number_of_cells,
                    np.nan, dtype=map_.dtype)
            map__[map_.index] = map_.values
            z[ok] = map__[cell_ix[ok]]
        z[~ok] = np.nan

        z = z.reshape(x.shape)
        surface = go.Surface(x=x, y=y, z=z, **surface_kwargs)

        if fig is None:
            fig = go.Figure(data=[surface], **figure_kwargs)
        else:
            fig.add_trace(surface, **figure_kwargs)

__all__ = [ 'Plotly' ]




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd
try:
    import polytope as pt
except ImportError:
    pt = None
import copy
from tramway.core.xyt import crop, reindex_trajectories
import tramway.core.analyses.auto as autosaving
from tramway.helper import *
import re
import itertools
from collections import defaultdict, OrderedDict


try:
    from tqdm import tqdm
except ImportError:
    pass
else:
    rc.__available_packages__.add('tqdm')


class SupportRegions(object):
    """
    A support region can be either a region of interest or the union of regions of interest.
    Major processing steps such as tessellation and inference may operate
    on support regions so that a region of interest simply acts as a window.

    This class offers a base implementation for :class:`UnitRegions` and :class:`GroupedRegions`.
    """
    __slots__ = ('gen_label','update_metadata','verbose')
    def __init__(self, region_label=None, update_metadata=None, verbose=True):
        self.__reset__()
        self.gen_label = region_label
        self.update_metadata = update_metadata
        self.verbose = verbose
    def __reset__(self):
        raise NotImplementedError('abstract method')
    def tessellate(self, r, analysis_tree, *args, **kwargs):
        if isinstance(r, str):
            if r == 'all':
                description = 'Tessellating the regions of interest'
            else:
                description = r
            any_new = False
            for r in self.iter_regions(description):
                if self.tessellate(r, analysis_tree, *args, **kwargs):
                    any_new = True
            return any_new
        #
        label = self.region_label(r)
        if label in analysis_tree:
            return False
        else:
            if self.verbose:
                print(label)
            trajectories = self.crop(r, analysis_tree.data)
            partition = tessellate(trajectories, *args, **kwargs)
            analysis_tree[label] = partition
            if self.update_metadata is not None:
                self.update_metadata(analysis_tree[label])
            return True
    def infer(self, r, analysis_tree, *args, **kwargs):
        if isinstance(r, str) and (r == 'all' or r not in analysis_tree):
            if r == 'all':
                description = 'Inferring dynamics parameters'
            else:
                description = r
            any_new = False
            for r in self.iter_regions(description):
                if self.infer(r, analysis_tree, *args, **kwargs):
                    any_new = True
            return any_new
        #
        if isinstance(r, str):
            label = r
        else:
            label = self.region_label(r)
        if label in analysis_tree:
            skip_interrupted = kwargs.pop('preserve_interrupted_inferences', False)
            try:
                maps = analysis_tree[label][kwargs['output_label']].data
            except KeyError: # either 'output_label' in kwargs or kwargs['output_label'] in analysis_tree
                pass
            else:
                try:
                    if skip_interrupted or maps.resolution.upper() != 'INTERRUPTED':
                        return False
                except AttributeError: # either resolution in maps or upper in maps.resolution
                    return False
            kwargs['input_label'] = label
            output_label = analysis_tree[label].autoindex(kwargs.get('output_label', None))
            if self.verbose:
                print('{} -- {}'.format(label, output_label))
            infer(analysis_tree, *args, **kwargs)
            if self.update_metadata is not None:
                self.update_metadata(analysis_tree[label][output_label])
            return True
        else:
            import warnings
            warnings.warn("no partition available for region '{}'".format(label))
        return False
    def reset_roi(self, r, analysis_tree, *args, **kwargs):
        del analysis_tree[self.region_label(r)]
    def __range__(self, n, desc=None):
        iterable = not isinstance(n, int)
        if iterable:
            iterator = n
            n = len(iterator)
        if self.verbose and rc.__user_interaction__ and rc.__has_package__('tqdm'):
            iternum = tqdm(range(n), desc=desc)
        else:
            iternum = range(n)
        if iterable:
            return zip(iternum, iterator)
        else:
            return iternum

class UnitRegions(SupportRegions):
    """
    Regions of interest are considered separately, independently of whether they overlap or not.
    """
    __slots__ = ('unit_region','_bw_comp',)
    def __init__(self, *args, **kwargs):
        SupportRegions.__init__(self, *args, **kwargs)
        self._bw_comp = False
    def __reset__(self):
        self.unit_region = OrderedDict()
    def add_collection(self, unit_regions, label=None):
        if label is None:
            label = ''
        self.unit_region[label] = unit_regions
    def __len__(self):
        return sum([0]+[ len(self.unit_region[r]) for r in self.unit_region ])
    def __contains__(self, r):
        return 0<=r and r<len(self)
    def __iter__(self):
        yield from range(len(self))
    def __getitem__(self, r):
        _r = r
        for coll in self.unit_region:
            rs = self.unit_region[coll]
            if _r<len(rs):
                return rs[_r]
            else:
                _r -= len(rs)
        raise IndexError('out of bounds: {}'.format(r))
    def __setitem__(self, r, bounds):
        _r = r
        for coll in self.unit_region:
            rs = self.unit_region[coll]
            if _r<len(rs):
                rs[_r] = bounds
                return
            else:
                _r -= len(rs)
        raise IndexError('out of bounds: {}'.format(r))
    def region_label(self, r):
        if not isinstance(r, int):
            assert not r[1:]
            r = r[0]
        for collection in self.unit_region:
            n = len(self.unit_region[collection])
            if n <= r:
                r -= n
            else:
                break
        if self._bw_comp:
            return self.gen_label(r, collection if collection else None)
        else:
            return self.gen_label({ collection if collection else '': [r] })
    @property
    def region_labels(self):
        labels = []
        for collection in self.unit_region:
            n = len(self.unit_region[collection])
            if not collection:
                collection = None
            for r in range(n):
                if self._bw_comp:
                    labels.append(self.gen_label(r, collection))
                else:
                    labels.append(self.gen_label({ collection if collection else '': [r] }))
        return labels
    def unit_to_region(self, u, collection=None):
        if collection is None:
            collection = ''
        r = 0
        for label in self.unit_region:
            if label == collection:
                return r+u
            else:
                r += len(self.unit_region[label])
        raise KeyError("no such roi collection: '{}'".format(collection))
    def region_to_units(self, r):
        """
        Returns a `dict` with a single key and single-element list value.

        For compatibility with : class:`GroupedRegions`.
        """
        _r = r
        for coll in self.unit_region:
            rs = self.unit_region[coll]
            if _r<len(rs):
                return { coll: [_r] }
            else:
                _r -= len(rs)
        raise IndexError('out of bounds: {}'.format(r))
    #def collection_range(self, collection_label):
    #    m = 0
    #    for label in self.unit_region:
    #        k = len(self.unit_region[label])
    #        if collection_label == label:
    #            n = m + k
    #            break
    #        else:
    #            m += k
    #    return m, n
    def iter_regions(self, desc=None):
        return self.__range__(len(self), desc)
    def crop(self, r, df, **kwargs):
        n_space_cols = len([ col for col in 'xyz' if col in df.columns ])
        for regions in self.unit_region.values():
            if len(regions) <= r:
                r -= len(regions)
            else:
                if isinstance(regions[r], (pt.Polytope, pt.Region)):
                    raise NotImplementedError
                else:
                    _min,_max = regions[r]
                    if n_space_cols < _min.size:
                        assert _min.size == n_space_cols + 1
                        df = df[(_min[-1] <= df['t']) & (df['t'] <= _max[-1])]
                        df = crop(df, np.r_[_min[:-1], _max[:-1]-_min[:-1]], **kwargs)
                    else:
                        df = crop(df, np.r_[_min,_max-_min], **kwargs)
                return df

class GroupedRegions(SupportRegions):
    """
    Overlapping regions of interest are pooled together and form a unique support region.

    Pooling disregards which collection a ROI pertains to.
    """
    def __reset__(self):
        self.unit_region = []
        self._unit_polytope = {}
        self.group = {}
        self.index = OrderedDict()
    def unit_bounding_box(self, i, exact_only=False, atleast_2d=False):
        r = self.unit_region[i]
        if isinstance(r, tuple):
            if atleast_2d:
                _min,_max = r
                if not _min.shape[1:]:
                    assert not _max.shape[1:]
                    r = _min[np.newaxis,:], _max[np.newaxis,:]
        elif isinstance(r, list):
            if atleast_2d:
                _min,_max = r
                if _min.shape[1:]:
                    assert _max.shape[1:]
                    r = tuple(r)
                else:
                    assert not _max.shape[1:]
                    r = _min[np.newaxis,:], _max[np.newaxis,:]
            else:
                r = tuple(r)
            self.unit_region[i] = r
        else:#isinstance(r, (pt.Polytope, pt.Region)):
            if exact_only:
                r = None, None
            elif atleast_2d:
                r = r.bounding_box
            else:
                _min,_max = r.bounding_box
                r = np.ravel(_min), np.ravel(_max)
        return r
    def unit_polytope(self, i):
        try:
            return self._unit_polytope[i]
        except KeyError:
            r = self.unit_region[i]
            if isinstance(r, pt.Region):
                pass
            else:
                if isinstance(r, (tuple, list)):
                    r = pt.box2poly(list(zip(*r)))
                    self._unit_polytope[i] = r
                r = pt.Region([r])
            return r
    @property
    def unit_bounding_boxes(self):
        return [ self.unit_bounding_box(i) for i in range(len(self.unit_region)) ]
    @property
    def unit_polytopes(self):
        return [ self.unit_polytope(i) for i in range(len(self.unit_region)) ]
    def adjacent(self, i, j):
        _min_i,_max_i = self.unit_bounding_box(i, True)
        _min_j,_max_j = self.unit_bounding_box(j, True)
        if _min_i is None or _min_j is None:
            _ri,_rj = self.unit_polytope(i), self.unit_polytope(j)
            return pt.is_adjacent(_ri,_rj)
        else:
            return np.all(_min_i<=_max_j) and np.all(_min_j<=_max_i)
    def add_collection(self, unit_regions, label=None):
        # check if already existing
        if label in self.index:
            if 1<len(self.index):
                raise RuntimeError('cannot overwrite a collection if other collections have already been defined')
            self.__reset__()
        #
        i0 = len(self.unit_region)
        self.unit_region += list(unit_regions)
        # first group overlapping unit regions in the collection
        current_index = max(self.group.keys())+1 if self.group else 0
        not_an_index = -1
        n = len(unit_regions)
        assignment = np.full(n, not_an_index, dtype=int)
        groups = dict()
        for i in range(n):
            region_i = unit_regions[i]
            if isinstance(region_i, pt.Polytope):
                region_i = pt.Region([region_i])
                _min_i = _max_i = None
            elif isinstance(region_i, pt.Region):
                _min_i = _max_i = None
            else:#if isinstance(region_i, (tuple, list)):
                _min_i, _max_i = region_i
                region_i = None
            group_with = set()
            if assignment[i] == not_an_index:
                group_index = current_index
                group = set([i0+i])
            else:
                group_index = assignment[i]
                group_with.add(group_index)
                group = groups[group_index]
                assert i0+i in group
            for j in range(i+1,n):
                if i0+j in group:
                    continue
                region_j = unit_regions[j]
                if isinstance(region_j, (pt.Polytope, pt.Region)):
                    if region_i is None:
                        region_i = pt.box2poly(list(zip(_min_i,_max_i)))
                        self._unit_polytope[i0+i] = region_i
                    i_and_j_are_adjacent = pt.is_adjacent(region_i, region_j)
                else:#if isinstance(region_j, (tuple, list)):
                    _min_j,_max_j = region_j
                    if _min_i is None:
                        region_j = pt.box2poly(list(zip(_min_j,_max_j)))
                        self._unit_polytope[i0+j] = region_j
                        i_and_j_are_adjacent = pt.is_adjacent(region_i, region_j)
                    else:
                        i_and_j_are_adjacent = np.all(_min_i<=_max_j) and np.all(_min_j<=_max_i)
                if i_and_j_are_adjacent:
                    if assignment[j]==not_an_index:
                        group.add(i0+j)
                    else:
                        other_group_index = assignment[j]
                        group_with.add(other_group_index)
                        group |= groups.pop(other_group_index)
            if group_with:
                group_index = min(group_with)
            else:
                current_index += 1
            groups[group_index] = group
            group = np.array(list(group)) - i0 # indices in `assignment`
            assignment[group] = group_index
        #
        # merge the new and existing groups together
        for g in list(groups.keys()):
            adjacent = set()
            for h in self.group:
                g_and_h_are_adjacent = False
                for i in groups[g]:
                    for j in self.group[h]:
                        if self.adjacent(i,j):
                            g_and_h_are_adjacent = True
                            break
                    if g_and_h_are_adjacent:
                        adjacent.add(h)
                        break
            if adjacent:
                h = min(adjacent)
                for i in adjacent - {h}:
                    self.group[h] |= self.group.pop(i)
                self.group[h] |= groups.pop(g)
                assignment[assignment==g] = h
        if groups:
            self.group.update(groups)
        #
        if label is None:
            label = ''
        self.index[label] = assignment
        #
        self.reverse_index = np.c_[
                np.repeat(np.arange(len(self.index)), [ len(self.index[s]) for s in self.index ]),
                np.concatenate([ np.arange(len(self.index[s])) for s in self.index ]) ]
    def unit_to_region(self, u, collection=None):
        if collection is None:
            collection = ''
        return self.index[collection][u]
    def region_to_units(self, r):
        coll_num = self.reverse_index[list(self.group[r])]
        units = defaultdict(list)
        for coll_ix, ix_in_coll in coll_num:
            label = self.collection_labels[coll_ix]
            units[label].append(ix_in_coll)
        return units
    @property
    def collection_labels(self):
        return list(self.index.keys())
    def region_label(self, r):
        return self.gen_label(self.region_to_units(r))
    @property
    def region_labels(self):
        return [ self.region_label(r) for r in self.group ]
    def __len__(self):
        return len(self.group)
    def __contains__(self, r):
        return r in self.group
    def __iter__(self):
        return iter(self.group)
    def __getitem__(self, r):
        return self.group[r]
    #def __setitem__(self, r, roi_set):
    #    self.group[r] = roi_set
    #def __delitem__(self, r):
    #    del self.group[r]
    def collection_range(self, collection_label):
        m = 0
        for label in self.index:
            k = len(self.index[label])
            if collection_label == label:
                n = m + k
                break
            else:
                m += k
        return m, n
    def iter_regions(self, desc=None):
        rs = list(self.group.keys())[::-1]
        done = 0
        for _ in self.__range__(len(self.unit_region), desc):
            if done == 0:
                r = rs.pop()
                unit_roi = self.region_to_units(r)
                done = sum([ len(roi) for roi in unit_roi.values() ])
                yield r
            done -= 1
    def crop(self, r, df, **kwargs):
        kwargs['preserve_index'] = True
        n_space_cols = len([ col for col in 'xyz' if col in df.columns ])
        loc_indices = set()
        df_r = None
        for u in self.group[r]:
            if isinstance(self.unit_region[u], (pt.Polytope, pt.Region)):
                raise NotImplementedError
            else:
                _min,_max = self.unit_region[u]
                if n_space_cols < _min.size:
                    assert _min.size == n_space_cols + 1
                    df_u = df[(_min[-1] <= df['t']) & (df['t'] <= _max[-1])]
                    df_u = crop(df_u, np.r_[_min[:-1], _max[:-1]-_min[:-1]],
                            **kwargs)
                else:
                    df_u = crop(df, np.r_[_min,_max-_min], **kwargs)
            if df_r is None:
                df_r = df_u
            else:
                df_r = pd.merge(df_r, df_u, how='outer')
        return reindex_trajectories(df_r.sort_values(by=['n','t']))


class RoiCollection(object):
    """
    """
    def __init__(self, roi, label=None, regions=None):
        self.label = label
        self.regions = regions
        self.regions.add_collection(roi, label)
    @property
    def bounding_box(self):
        try:
            return self.regions.unit_region[self.label]
        except TypeError:
            m,n = self.regions.collection_range(self.label)
            return self.regions.unit_bounding_boxes[m:n]
    def __len__(self):
        raise NotImplementedError
        # TODO: this collection only?
        return len(self.regions.unit_region)
    def subset_index(self, r):
        return self.regions.unit_to_region(r, self.label)
    def subset_label(self, r=None, s=None):
        if s is None:
            s = self.subset_index(r)
        return self.regions.region_label(s)
    def get_subset(self, r=None, s=None):
        if s is None:
            s = self.subset_index(r)
        try:
            return self.regions.group[s]
        except AttributeError:
            return [s]
    def roi_label(self, r):
        try:
            return self.regions.unit_region_label(r, self.label)
        except AttributeError:
            return self.regions.gen_label({self.label: [r]})
    def get_subtree(self, i, analysis_tree):
        label = self.subset_label(i)
        if label in analysis_tree:
            return analysis_tree[label]
    def overlaps(self, i):
        return 1 < len(self.get_subset(i))
    def get_map(self, i, analysis_tree, map_label, full=False):
        """
        returns the `Map` object for unit roi *i*.
        """
        label = self.subset_label(i)
        try:
            subtree = analysis_tree[label]
            maps = subtree[map_label].data
        except KeyError:
            return None
        if not full and self.overlaps(i):
            tessellation = subtree.data.tessellation
            try:
                segments = len(tessellation.time_lattice)
            except AttributeError:
                segments = None
            else:
                tessellation = tessellation.spatial_mesh
            coords = tessellation.cell_centers
            dim = coords.shape[1]
            _min,_max = self.bounding_box[i]
            inside = np.all((_min[np.newaxis,:dim]<coords) & (coords<_max[np.newaxis,:dim]),axis=1)
            if segments:
                inside = np.tile(inside, segments)
            maps.maps = maps.maps[inside[maps.maps.index]]
        return maps
    def get_cells(self, i, analysis_tree):
        """
        returns the `Partition` object for unit roi *i*.
        """
        label = self.subset_label(i)
        cells = analysis_tree[label].data
        tessellation = cells.tessellation
        try:
            tessellation = tessellation.spatial_mesh
        except AttributeError:
            pass
        if self.overlaps(i):
            coords = tessellation.cell_centers
            dim = coords.shape[1]
            _min,_max = self.bounding_box[i]
            inner_cell = np.all((_min[np.newaxis,:dim]<coords) & (coords<_max[np.newaxis,:dim]),axis=1)
        else:
            inner_cell = np.ones(tessellation.number_of_cells, dtype=bool)
        return cells, inner_cell
    def get_tessellation(self, i, analysis_tree):
        """
        returns the `Tessellation` object for unit roi *i*,
        together with a boolean array so that the *k*-th element is ``True``
        if microdomain *k* lies within roi *i*.
        """
        cells, inner_cell = self.get_cells(i, analysis_tree)
        return cells.tessellation, inner_cell
    def cell_plot(self, i, analysis_tree, decorate=True, **kwargs):
        """
        plots the `Partition` for unit roi *i*.
        """
        if isinstance(i, str):
            label = i
            title = kwargs.pop('title', label)
        else:
            label = self.subset_label(i)
            title = kwargs.pop('title', self.roi_label(i))
        if 'delaunay' not in kwargs:
            # anticipate future changes in `helper.tessellation.cell_plot`
            voronoi = kwargs.pop('voronoi', dict(centroid_style=None))
            kwargs['voronoi'] = voronoi
        if 'aspect' not in kwargs:
            kwargs['aspect'] = 'equal'
        plot_bb = decorate and self.overlaps(i)
        if True:#plot_bb:
            kwargs['show'] = False
        #
        cell_plot(analysis_tree, label=label, title=title, **kwargs)
        #
        if decorate:
            if 'axes' in kwargs:
                ax = kwargs['axes']
            else:
                import matplotlib.pyplot as plt
                ax = plt.gca()
            _min,_max = self.bounding_box[i]
            x0, y0, x1, y1 = _min[0], _min[1], _max[0], _max[1]
            xl, yl = ax.get_xlim(), ax.get_ylim()
            xc, yc = .5 * (x0 + x1), .5 * (y0 + y1)
            if not plot_bb:
                x0, y0, x1, y1 = xl[0], yl[0], xl[1], yl[1]
            ax.plot(
                    [x0, x1, np.nan, xc, xc],
                    [yc, yc, np.nan, y0, y1],
                    color='k', linestyle='--', alpha=1, linewidth=1)
        if plot_bb:
            from matplotlib.collections import PatchCollection
            from matplotlib.patches import Rectangle
            masks = []
            if xl[0]<x0:
                masks.append(Rectangle((xl[0], yl[0]), x0-xl[0], yl[1]-yl[0]))
            if x1<xl[1]:
                masks.append(Rectangle((x1, yl[0]), xl[1]-x1, yl[1]-yl[0]))
            if yl[0]<y0:
                masks.append(Rectangle((x0, yl[0]), x1-x0, y0-yl[0]))
            if y1<yl[1]:
                masks.append(Rectangle((x0, y1), x1-x0, yl[1]-y1))
            pc = PatchCollection(masks, facecolor='k', alpha=.1, edgecolor=None)
            ax.add_collection(pc)
            #
            ax.plot(
                    [x0,x0,x1,x1,x0],
                    [y0,y1,y1,y0,y0],
                    color='k', linestyle='-', alpha=1, linewidth=1)
            ax.set_xlim(xl)
            ax.set_ylim(yl)
    def map_plot(self, i, analysis_tree, map_label, decorate=True, **kwargs):
        """
        plots the `Maps` for unit roi *i*.
        """
        if isinstance(i, str):
            label = i
            title = kwargs.pop('title', label)
        else:
            label = self.subset_label(i)
            title = kwargs.pop('title', self.roi_label(i))
        if 'aspect' not in kwargs:
            kwargs['aspect'] = 'equal'
        kwargs['show'] = False
        plot_bb = decorate and self.overlaps(i)
        #
        map_plot(analysis_tree, label=(label,map_label), title=title, **kwargs)
        #
        if decorate:
            _min,_max = self.bounding_box[i]
            x0, y0, x1, y1 = _min[0], _min[1], _max[0], _max[1]
            xc, yc = .5 * (x0 + x1), .5 * (y0 + y1)
            if 'axes' in kwargs:
                ax = kwargs['axes']
            else:
                import matplotlib.pyplot as plt
                ax = plt.gca()
            ax.plot(xc, yc, 'r+')
        if plot_bb:
            from matplotlib.collections import PatchCollection
            from matplotlib.patches import Rectangle
            xl, yl = ax.get_xlim(), ax.get_ylim()
            #
            masks = []
            if xl[0]<x0:
                masks.append(Rectangle((xl[0], yl[0]), x0-xl[0], yl[1]-yl[0]))
            if x1<xl[1]:
                masks.append(Rectangle((x1, yl[0]), xl[1]-x1, yl[1]-yl[0]))
            if yl[0]<y0:
                masks.append(Rectangle((x0, yl[0]), x1-x0, y0-yl[0]))
            if y1<yl[1]:
                masks.append(Rectangle((x0, y1), x1-x0, yl[1]-y1))
            pc = PatchCollection(masks, facecolor='k', alpha=.1, edgecolor=None)
            ax.add_collection(pc)
            #
            ax.plot(
                    [x0,x0,x1,x1,x0],
                    [y0,y1,y1,y0,y0],
                    'k-', linewidth=1)
            ax.set_xlim(xl)
            ax.set_ylim(yl)
    def reset_roi(self, i, analysis_tree):
        self.regions.reset_roi(self.get_subset_index(i), analysis_tree)

class RoiCollections(AutosaveCapable):
    """
    collection of all collections of roi.

    manages the analysis tree for the spt data.
    Modifications can be automatically saved into an *.rwa* file.
    """
    def __init__(self, group_overlapping_roi=False, rwa_file=None, autosave=True, metadata=None, verbose=True, _bw_comp=False):
        AutosaveCapable.__init__(self, rwa_file, autosave)
        label = self.roi_label
        kwargs = dict()
        if group_overlapping_roi:
            Regions = GroupedRegions
        else:
            Regions = UnitRegions
            if _bw_comp:
                label = self.single_roi_label
        self.regions = Regions(
                region_label=label,
                update_metadata=metadata,
                verbose=verbose)
        if group_overlapping_roi:
            self.regions.unit_region_label = self.single_roi_label
        elif _bw_comp:
            self.regions._bw_comp = True
        self.collections = {}
    def __len__(self):
        return len(self.collections)
    def __contains__(self, label):
        return label in self.collections
    def __iter__(self):
        return iter(self.collections)
    def __setitem__(self, label, roi):
        self.collections[label] = roi if isinstance(roi, RoiCollection) else RoiCollection(roi, label, self.regions)
    def __getitem__(self, label):
        return self.collections[label]
    def __delitem__(self, label):
        del self.collections[label]
    @property
    def verbose(self):
        return self.regions.verbose
    @verbose.setter
    def verbose(self, v):
        self.regions.verbose = v
    @property
    def numeric_format(self):
        return '{:0>3d}'
    def roi_label(self, coll_num):
        label = []
        for coll in coll_num:
            num = sorted(coll_num[coll])
            num_label = '-'.join([ self.numeric_format.format(i) for i in num ])
            if coll:
                label.append( ' '.join((coll, num_label)) )
            else:
                label.append( 'roi'+num_label )
        return ' - '.join(label)
    def single_roi_label(self, i, collection_label=None):
        """ for figure titles """
        num_label = self.numeric_format.format(i) if isinstance(i, int) else i
        if collection_label:
            return '{} roi {}'.format(collection_label, num_label)
        else:
            return 'roi'+num_label
    @property
    def roi_labels(self):
        return self.regions.region_labels
    @property
    def tessellate_desc(self):
        return 'Tessellating the regions of interest'
    def tessellate(self, analysis_tree, *args, **kwargs):
        self.regions.tessellate(self.tessellate_desc, analysis_tree, *args, **kwargs)
    @property
    def infer_desc(self):
        return 'Inferring dynamics parameters'
    def infer(self, analysis_tree, *args, **kwargs):
        self.regions.infer(self.infer_desc, analysis_tree, *args, **kwargs)

    def reset(self):
        self.regions.__reset__()
        self.collections = {}


#from .base import Helper, Analyses
#from tramway.helper.tessellation import Tessellate, Partition, Voronoi
import scipy.sparse as sparse
import itertools

class RoiHelper(Helper):
    """
    maintains an extra data artefact in the *.rwa* file to store/load
    roi definitions.
    """
    def __init__(self, input_data, roi=None,
            meta_label='all %Sroi', meta_label_sep=' ',
            rwa_file=None, autosave=True, verbose=True,
            group_overlapping_roi=True, _bw_comp=False):
        Helper.__init__(self)
        input_data = Tessellate.prepare_data(self, input_data)

        if rwa_file is None and isinstance(input_data, Analyses):
            rwa_file = self.input_file
            if isinstance(rwa_file, list):
                assert not rwa_file[1:]
                rwa_file = rwa_file[0]

        analyses = self.analyses
        self.analyses = autosaving.Analyses(rwa_file=rwa_file, autosave=autosave)
        self.analyses.analyses = analyses

        self.meta_label_pattern = meta_label
        self.meta_label_sep = meta_label_sep

        #self.collections = RoiCollections(group_overlapping_roi,
        #        rwa_file, autosave, self.add_metadata, verbose)
        self.collections = RoiCollections(autosave=False, metadata=self.add_metadata,
                group_overlapping_roi=group_overlapping_roi, _bw_comp=_bw_comp)

        if roi is None:
            for coll_label, meta_label in self.get_meta_labels().items():
                self.collections[coll_label] = self.get_bounding_boxes(meta_label=meta_label)
        elif isinstance(roi, dict):
            for label in roi:
                self.set_bounding_boxes(roi[label], collection_label=label)
        else:
            self.set_bounding_boxes(roi)
            self.collections[''] = roi

        self._extra_artefacts = {}

    def get_meta_labels(self):
        pattern = self.meta_label_pattern.replace('%s', '(?P<collection>.*)')
        pattern = pattern.replace('%S', '((?P<collection>.+){})?'.format(self.meta_label_sep))
        labels = {}
        for label in self.analyses.labels:
            match = re.fullmatch(pattern, label)
            if match is not None:
                collection = match.group('collection')
                if collection is None:
                    collection = ''
                labels[collection] = label
        return labels

    def to_meta_label(self, collection_label):
        if collection_label:
            meta_label = self.meta_label_pattern.replace('%S', collection_label+self.meta_label_sep)
        else:
            collection_label = ''
            meta_label = self.meta_label_pattern.replace('%S', '')
        meta_label = meta_label.replace('%s', collection_label)
        return meta_label

    def set_bounding_boxes(self, roi, roi_size=None, collection_label=None):

        trajectories = self.analyses.data

        space_cols = [ col for col in 'xyz' if col in trajectories.columns ]
        def _space(_xyt):
            _xy = _xyt[:-1] if len(space_cols) < _xyt.size else _xyt
            return _xy

        roi_centers = np.stack([ _space((_min+_max)/2) for _min,_max in roi ], axis=0)
        roi_adjacency = sparse.coo_matrix(([],([],[])), shape=(len(roi),len(roi)), dtype=bool)
        # 2D only
        roi_vertices = np.vstack(list(itertools.chain(*[
            [ [_min[0],_min[1]], [_min[0],_max[1]], [_max[0],_max[1]],  [_max[0],_min[1]] ] for _min,_max in roi ])))
        roi_cell_vertices = list(np.reshape(np.arange(len(roi_vertices)), (len(roi),-1)))
        roi_vertex_adjacency_row = np.concatenate([ 4*i + np.array([0, 0, 1, 2]) for i in range(len(roi)) ])
        roi_vertex_adjacency_col = np.concatenate([ 4*i + np.array([1, 3, 2, 3]) for i in range(len(roi)) ])
        _n_edges = roi_vertex_adjacency_row.size
        roi_vertex_adjacency = sparse.csr_matrix((np.ones((2*_n_edges,), dtype=bool),
               (np.r_[roi_vertex_adjacency_row, roi_vertex_adjacency_col],
                np.r_[roi_vertex_adjacency_col, roi_vertex_adjacency_row])), shape=(_n_edges, _n_edges))

        if roi_size is None:
            # assume same shape
            roi_size = roi_vertices[2,0] - roi_vertices[1,0]

        #roi_polytopes = [ pt.box2poly([[_min[0],_max[0]],[_min[1],_max[1]]]) for _min,_max in roi ]
        roi_index = []
        #for i, x in enumerate(trajectories[['x','y']].values):
        #    for j, p in enumerate(roi_polytopes):
        #        if x in p:
        #            roi_index.append((i,j))
        #roi_index = tuple([ np.array(a) for a in zip(*roi_index) ])
        xy = xyt = None
        I = []
        for j, (_min,_max) in enumerate(roi):
            if len(space_cols) < _min.size:
                if xyt is None:
                    xyt = trajectories[space_cols+['t']].values
                pt = xyt
            else:
                if xy is None:
                    xy = trajectories[space_cols].values
                pt = xy
            i, = np.nonzero(np.all((_min[np.newaxis,:] <= pt) & (pt <= _max[np.newaxis,:]), axis=1))
            I.append(i)
        J = np.repeat(np.arange(len(I)), [ len(i) for i in I])
        I = np.concatenate(I)
        roi_index = I,J

        roi_mesh = Voronoi() # not really a Voronoi diagram...
        roi_mesh._preprocess(trajectories[space_cols])
        roi_mesh.cell_centers = roi_centers
        roi_mesh.cell_adjacency = roi_adjacency
        roi_mesh.vertices = roi_vertices
        roi_mesh.cell_vertices = roi_cell_vertices
        roi_mesh.vertex_adjacency = roi_vertex_adjacency
        roi_mesh.cell_volume = np.full(len(roi), roi_size * roi_size)

        # TODO: wrap the spatial mesh into a TimeLattice object to store the time segment

        roi_partition = Partition(trajectories, roi_mesh)
        roi_partition.cell_index = roi_index

        #from tramway.tessellation.nesting import NestedTessellations
        #from tramway.tessellation.hexagon import HexagonalMesh
        #roi_partitions = Partition(trajectories,
        #            NestedTessellations(parent=roi_partition, factory=HexagonalMesh, ref_distance=.01))

        meta_label = self.to_meta_label(collection_label)
        self.analyses[meta_label] = roi_partition
        self.add_metadata(self.analyses[meta_label])

        if collection_label is None:
            collection_label = ''
        self.collections[collection_label] = roi

    def get_bounding_boxes(self, collection_label=None, meta_label=None):
        if meta_label is None:
            if collection_label is None:
                collection_label = ''
            return self.collections[collection_label].bounding_box

        roi_mesh = self.get_global_partition(meta_label=meta_label).tessellation
        roi = []
        for vertex_indices in roi_mesh.cell_vertices:
            assert len(vertex_indices) == 4
            bottom_left = roi_mesh.vertices[vertex_indices[0]]
            top_right = roi_mesh.vertices[vertex_indices[2]]
            roi.append((bottom_left, top_right))
        return roi

    def get_global_partition(self, collection_label=None, meta_label=None):
        if meta_label is None:
            meta_label = self.to_meta_label(collection_label)
        return self.analyses[meta_label].data

    @property
    def roi_labels(self):
        return self.collections.roi_labels

    def tessellate(self, *args, **kwargs):
        with self.analyses.autosaving(kwargs.pop('autosave',None)):
            return self.collections.tessellate(self.analyses, *args, **kwargs)

    def infer(self, *args, **kwargs):
        with self.analyses.autosaving(kwargs.pop('autosave',None)):
            return self.collections.infer(self.analyses, *args, **kwargs)

    def cell_plot(self, i, collection_label='', **kwargs):
        return self.collections[collection_label].cell_plot(i, self.analyses, **kwargs)

    def map_plot(self, i, map_label, collection_label='', **kwargs):
        return self.collections[collection_label].map_plot(i, self.analyses, map_label, **kwargs)

    def get_tessellation(self, i, collection_label=''):
        return self.collections[collection_label].get_tessellation(i, self.analyses)

    def get_map(self, i, map_label, full=False, collection_label=''):
        return self.collections[collection_label].get_map(i, self.analyses, map_label, full)

    def reset_roi(self, i, collection_label=''):
        self.collections[collection_label].reset_roi(i)

    def save_analyses(self, output_file=None, verbose=None, force=None, **kwargs):
        if output_file is not None:
            raise NotImplementedError
        self.analyses.save(out_of_context=True)
        return
        # deprecated
        if output_file is None:
            output_file = self.collections.rwa_file
        if verbose is None:
            verbose = self.collections.verbose
        Helper.save_analyses(self, output_file, verbose, force, **kwargs)

    def set_base_grid(self, label, grid):
        label = 'tramway.roi.RoiHelper.basegrid:'+label
        self.analyses[label] = grid
        self.add_metadata(self.analyses[label])

    def get_base_grid(self, label, from_root=False):
        label = 'tramway.roi.RoiHelper.basegrid:'+label
        if from_root:
            try:
                artefact = self._extra_artefacts[label]
            except KeyError:
                artefact = None
                if self.collections.rwa_file:
                    from rwa import HDF5Store, lazyvalue
                    f = HDF5Store(self.collections.rwa_file, 'r')
                    try:
                        artefact = lazyvalue(f.peek(label))
                    except KeyError:
                        pass
                    else:
                        self._extra_artefacts[label] = artefact
                    finally:
                        f.close()
        else:
            artefact = self.analyses[label].data
        return artefact

    def reset(self):
        self.collections.reset()


__all__ = [ 'SupportRegions', 'UnitRegions', 'GroupedRegions', 'RoiCollection', 'RoiCollections', 'RoiHelper' ]





from tramway.core.analyses.browser import AnalysisBrowser
from tramway.core.exceptions import RWAFileException
from tramway.tessellation.time import TimeLattice
from .map import *
import sys
import os
import numpy as np
import pandas as pd
import time
import warnings
import traceback
from collections import defaultdict
from bokeh.plotting import curdoc, figure
from bokeh.models import Select, Slider, CheckboxGroup, TextInput, Button, Paragraph
from bokeh.layouts import row, column


def short(path):
    """ Returns the basename with no file extension. """
    return os.path.splitext(os.path.basename(path))[0]

class Model(object):
    """
    Stores the analyzer browsing state.
    """
    def __init__(self, analyzer):
        self.analyzer = analyzer
        self.spt_data_sources = [ short(f.source) for f in analyzer.spt_data ]
        self.sampling_labels = []
        self.mapping_labels = []
        self.current_spt_data = None
        self.current_analyses = None
        self.current_sampling = None
        self.current_mapping = None
        self.current_feature = None
        self._clim = None
    def select_spt_data(self, source_name):
        if self.current_spt_data is not None:
            self.release_spt_data()
        for name, f in zip(self.spt_data_sources, self.analyzer.spt_data):
            if name == source_name:
                break
        if name != source_name:
            raise KeyError('source not found')
        self.current_spt_data = f
        self.current_analyses = AnalysisBrowser(f.analyses)
        return f
    def release_spt_data(self):
        self.sampling_labels = []
        self.mapping_labels = []
        self.current_spt_data = None
        self.current_analyses = None
        self.current_sampling = None
        self.current_mapping = None
        self.current_feature = None
    def release_spt_data(self):
        pass
    def select_sampling(self, sampling_label):
        if self.current_sampling is not None:
            self.release_sampling()
        self.current_analyses.select_child(sampling_label)
        self.current_sampling = self.current_analyses.artefact
        return self.current_sampling
    def release_sampling(self, release_feature=True):
        if self.current_mapping is not None:
            self.release_mapping(release_feature)
        self.mapping_labels = None
        self.current_analyses.select_parent()
        self.current_sampling = None
    def select_mapping(self, mapping_label):
        if self.current_mapping is not None:
            self.release_mapping()
        self.current_analyses.select_child(mapping_label)
        self.current_mapping = self.current_analyses.artefact
        if self.has_time_segments():
            t = self.current_sampling.tessellation
            m = self.current_mapping
            self.clim = {}
            for feature in m.features:
                _m = m[feature].values
                if _m.shape[1:]:
                    if 1<_m.shape[1]:
                        _m = np.sqrt(np.sum(_m*_m, axis=1))
                    else:
                        _m = _m.flatten()
                self.clim[feature] = [_m.min(), _m.max()]
            self.current_mapping = { feature: t.split_segments(m[feature]) for feature in m.features }
        return self.current_mapping
    def release_mapping(self, release_feature=True):
        if self.current_feature is not None and release_feature:
            self.release_feature()
        self.current_analyses.select_parent()
        self.current_mapping = None
    @property
    def features(self):
        assert self.current_mapping is not None
        if self.has_time_segments():
            return list(self.current_mapping.keys())
        else:
            return list(self.current_mapping.features)
    def select_feature(self, feature):
        self.current_feature = feature
    def release_feature(self):
        self.current_feature = None
    def has_time_segments(self):
        assert self.current_sampling is not None
        return isinstance(self.current_sampling.tessellation, TimeLattice)
    def n_time_segments(self):
        assert self.has_time_segments()
        return len(self.current_sampling.tessellation.time_lattice)
    @property
    def clim(self):
        if self._clim is None:
            clim = self.analyzer.browser.clim
        else:
            clim = self._clim
        if not isinstance(clim, dict):
            _clim = clim
            clim = defaultdict(lambda: _clim)
        return clim
    @clim.setter
    def clim(self, clim):
        if self.analyzer.browser.clim is None:
            self._clim = clim
        
class Controller(object):
    """
    Makes and manages the browser view.

    Note: time browsing is not supported yet.
    """
    def __init__(self, model, side_panel=None, webdriver=None, **map_kwargs):
        """
        Arguments:

            model (Model): analyzer browsing state.

            side_panel (bool): show the side panel with experimental export features.

            webdriver (type): *selenium* webdriver (subclass of
                :class:`selenium.webdriver.remote.webdriver.WebDriver`)
                for exporting figures.

        """
        self.model = model
        options = model.spt_data_sources
        self.source_dropdown = Select(title='Data source:', options=['None'] + options if options[1:] else options)
        self.sampling_dropdown = Select(title='Sampling:', disabled=True)
        self.mapping_dropdown = Select(title='Mapping:', disabled=True)
        self.feature_dropdown = Select(title='Feature:', disabled=True)
        self.source_dropdown.on_change('value', lambda attr, old, new: self.load_source(new))
        self.sampling_dropdown.on_change('value', lambda attr, old, new: self.load_sampling(new))
        self.mapping_dropdown.on_change('value', lambda attr, old, new: self.load_mapping(new))
        self.feature_dropdown.on_change('value', lambda attr, old, new: self.load_feature(new))
        self.trajectories_kwargs = dict(color='r', line_width=.5, line_alpha=.5, loc_alpha=.1, loc_size=6)
        self.show_side_panel = webdriver is not None if side_panel is None else side_panel
        self.selenium_webdriver = webdriver
        self.figure_export_width = self.figure_export_height = None
        # units may not be displayed on the colorbar, depending on bokeh
        # version
        self.map_kwargs = dict(unit='std')
        self.map_kwargs.update(map_kwargs)
    def load_source(self, source_name):
        if self.model.current_spt_data is not None:
            self.unload_source()
        if source_name == 'None':
            #self.unload_source()
            return
        try:
            self.model.select_spt_data(source_name)
        except RWAFileException:
            import traceback
            traceback.print_exc()
            self.source_dropdown.options = [ src for src in self.source_dropdown.options if src != source_name ]
            if not (self.source_dropdown.options and (self.source_dropdown.options[1:] or self.source_dropdown.options[0] != 'None')):
                self.source_dropdown.disabled = True
            return
        self.model.sampling_labels = list(self.model.current_analyses.labels())
        options = self.model.sampling_labels
        self.sampling_dropdown.options = ['None'] + options if options[1:] else options
        self.sampling_dropdown.disabled = False
        if not self.model.sampling_labels[1:]:
            if not self.model.sampling_labels:
                raise ValueError('no sampling available; please process the data first')
            self.load_sampling(self.model.sampling_labels[0])
    def unload_source(self):
        if self.model.current_sampling is not None:
            self.unload_sampling()
        self.sampling_dropdown.disabled = True
        self.sampling_dropdown.options = []
        self.model.sampling_labels = None
        self.model.release_spt_data()
    def load_sampling(self, sampling_label):
        if self.model.current_sampling is not None:
            self.unload_sampling()
        if sampling_label == 'None':
            return
        self.model.select_sampling(sampling_label)
        self.model.mapping_labels = list(self.model.current_analyses.labels())
        options = self.model.mapping_labels
        self.mapping_dropdown.options = ['None'] + options if options[1:] else options
        self.mapping_dropdown.disabled = False
        # time segmentation support
        if self.model.has_time_segments():
            self.time_slider.update(start=1, end=self.model.n_time_segments())
        #
        if not self.model.mapping_labels[1:]:
            self.load_mapping(self.model.mapping_labels[0])
    def unload_sampling(self):
        if self.model.current_mapping is not None:
            self.unload_mapping()
        self.mapping_dropdown.disabled = True
        self.mapping_dropdown.options = []
        self.model.release_sampling()
    def load_mapping(self, mapping_label):
        if self.model.current_mapping is not None:
            self.unload_mapping()
        if mapping_label == 'None':
            return
        self.model.select_mapping(mapping_label)
        options = self.model.features
        self.feature_dropdown.options = ['None'] + options if options[1:] else options
        self.feature_dropdown.disabled = False
        if not self.model.features[1:]:
            self.load_feature(self.model.features[0])
    def unload_mapping(self):
        if self.model.current_feature is not None:
            self.unload_feature()
        self.unset_export_status('data')
        self.feature_dropdown.disabled = True
        self.feature_dropdown.options = []
        self.model.release_mapping()
    def load_feature(self, feature):
        _curdoc = curdoc()
        _curdoc.hold()
        try:
            if feature == 'None':
                self.unload_feature()
                return
            self.unset_export_status('figure')
            self.model.select_feature(feature)
            if self.model.has_time_segments():
                self.enable_time_view()
            self.draw_map(feature)
            self.draw_trajectories()
            self.enable_space_view()
            self.enable_side_panel()
        finally:
            _curdoc.unhold()
    def unload_feature(self):
        self.disable_side_panel()
        self.unset_export_status('figure')
        self.disable_time_view()
        self.disable_space_view()
        if self.main_figure.renderers:
            self.main_figure.renderers = []
        if self.colorbar_figure.renderers:
            print(self.colorbar_figure.renderers)
            self.colorbar_figure.renderers = []
        self.feature_dropdown.value = 'None'
    def refresh_map(self):
        feature = self.model.current_feature
        if feature is None or feature == 'None':
            return
        _curdoc = curdoc()
        _curdoc.hold()
        try:
            self.unset_export_status('figure')
            self.draw_map(feature)
            self.draw_trajectories()
        finally:
            _curdoc.unhold()
    def make_main_view(self):
        """
        Makes the main view `browse_maps` adds as document root.
        """
        menu_view = row([self.source_dropdown, self.sampling_dropdown, self.mapping_dropdown, self.feature_dropdown])
        main_view = [menu_view]
        time_view = self.make_time_view()
        main_view.append(time_view)
        space_view = self.make_space_view()
        side_panel = self.make_side_panel()
        main_view.append(row(space_view, side_panel))
        return column(main_view)
    def make_time_view(self):
        self.time_slider = Slider(disabled=True, start=0, end=1, step=1)
        self.time_slider.on_change('value_throttled', lambda attr, old, new: self.refresh_map())
        return self.time_slider
    def make_space_view(self):
        self.main_figure = f = figure(disabled=True, toolbar_location=None,
                match_aspect=True, tools='pan, wheel_zoom, reset')
        self.main_figure.toolbar.active_drag = None
        f.background_fill_color = f.border_fill_color = None
        self.colorbar_figure = f = figure(disabled=True, toolbar_location=None,
                min_border=0, outline_line_color=None, title_location='right', plot_width=112)
        f.background_fill_color = f.border_fill_color = None
        f.title.align = 'center'
        add_colorbar(self.main_figure, colorbar_figure=f,
                colormap=self.model.analyzer.browser.colormap)
        #f.visible = False
        self.visibility_controls = CheckboxGroup(disabled=True, labels=['Localizations','Trajectories','Hide map'], active=[])
        def _update(attr, old, new):
            if 0 in old and 0 not in new:
                self.set_localization_visibility(False)
            elif 0 not in old and 0 in new:
                self.set_localization_visibility(True)
            if 1 in old and 1 not in new:
                self.set_trajectory_visibility(False)
            elif 1 not in old and 1 in new:
                self.set_trajectory_visibility(True)
            if 2 in old and 2 not in new:
                self.set_map_visibility(True)
            elif 2 not in old and 2 in new:
                self.set_map_visibility(False)
        self.visibility_controls.on_change('active', _update)
        return row(self.main_figure, self.colorbar_figure, self.visibility_controls)
    def disable_space_view(self):
        self.main_figure.disabled = True
        self.colorbar_figure.disabled = True
        self.visibility_controls.disabled = True
    def enable_space_view(self):
        #self.colorbar_figure.visible = True
        self.main_figure.disabled = False
        self.colorbar_figure.disabled = False
        self.visibility_controls.disabled = False
    def disable_time_view(self):
        self.time_slider.disabled = True
    def enable_time_view(self):
        self.time_slider.disabled = False
        assert 0<self.time_slider.start
        self.time_slider.value = 1
    def draw_map(self, feature):
        kwargs = dict(self.map_kwargs)
        if kwargs.get('unit', None) == 'std':
            # standard units are defined at multiple locations:
            # * tramway.plot.bokeh.analyzer.Controller.draw_map
            # * tramway.helper.inference.map_plot
            # * tramway.analyzer.mapper.mpl.Mpl.clabel
            unit = dict(
                    diffusivity='µm²/s',
                    potential='kT',
                    force='log. kT/µm',
                    drift='µm/s',
                    ) # LaTeX is not supported
            kwargs['unit'] = unit.get(feature, None)
        if self.model.analyzer.browser.colormap is not None:
            kwargs['colormap'] = self.model.analyzer.browser.colormap
        if self.main_figure.renderers:
            self.main_figure.renderers = []
        try:
            self.visibility_controls.active = [ pos for pos in self.visibility_controls.active if pos != 2 ]
        except ValueError:
            pass
        _cells = self.model.current_sampling
        _map = self.model.current_mapping[feature]
        _vector_map = None
        if self.model.has_time_segments():
            _cells = _cells.tessellation.split_segments(_cells)
            _seg = self.time_slider.value
            if _seg is None:
                warnings.warn('could not read time slider value', RuntimeWarning)
                _seg = 0
            else:
                _seg -= 1
            _cells = _cells[_seg]
            _map = _map[_seg]
            kwargs['clim'] = self.model.clim[feature]
        elif self.model.clim is not None:
            kwargs['clim'] = self.model.clim[feature]
        if temperature:
            # convert from µm^-1 (inferenceMAP kT) to true kT
            _map /= 6.950348e-5 * temperature
        if _map.shape[1] == 2:
            _vector_map = _map
            if feature == 'force':
                _map = _map.pow(2).sum(1).apply(np.log)*.5
            else:
                _map = _map.pow(2).sum(1).apply(np.sqrt)
        self.map_glyphs = scalar_map_2d(_cells, _map,
                figure=self.main_figure, colorbar_figure=self.colorbar_figure, **kwargs)
        if _vector_map is not None:
            self.map_glyphs.append(
                    field_map_2d(_cells, _vector_map,
                            figure=self.main_figure, inferencemap=True))
        elif _map.shape[1] != 1:
            raise NotImplementedError('neither a scalar map nor a 2D-vector map')
    def draw_trajectories(self):
        sampling = self.model.current_sampling
        if self.model.has_time_segments():
            try:
                seg = self.time_slider.value-1
            except TypeError:
                pass
            else:
                sampling = sampling.tessellation.split_segments(sampling)[seg]
        traj_handles = plot_trajectories(sampling.points,
                figure=self.main_figure, **self.trajectories_kwargs)
        self.trajectory_glyphs = traj_handles[0::2]
        self.location_glyphs = traj_handles[1::2]
        if 0 not in self.visibility_controls.active:
            self.set_localization_visibility(False)
        if 1 not in self.visibility_controls.active:
            self.set_trajectory_visibility(False)
    def set_localization_visibility(self, b):
        for handle in self.location_glyphs:
            handle.visible = b
    def set_trajectory_visibility(self, b):
        for handle in self.trajectory_glyphs:
            handle.visible = b
    def set_map_visibility(self, b):
        for handle in self.map_glyphs:
            handle.visible = b
    def enable_side_panel(self):
        if not self.show_side_panel:
            return
        self.export_file_input.disabled = False
        self.export_file_input.value = '_'.join([ part for part in (self.source_dropdown.value, self.sampling_dropdown.value, self.mapping_dropdown.value) if part ])
    def disable_side_panel(self):
        if not self.show_side_panel:
            return
        self.export_file_input.disabled = True
        self.export_file_input.value = ''
        self.figure_export_button.disabled = True
        self.data_export_button.disabled = True
    def make_side_panel(self):
        self.export_file_input = TextInput(disabled=True, value='', title='append any of: .png, .svg, .txt, .csv')
        self.figure_export_button = Button(disabled=True, label='Export figure', button_type='success')
        self.data_export_button = Button(disabled=True, label='Export data', button_type='success')
        def _update_buttons(attr, old, new):
            if new.endswith('.png') or new.endswith('.svg'):
                self.unset_export_status('figure')
                self.figure_export_button.disabled = False
            else:
                self.figure_export_button.disabled = True
            if new.endswith('.txt') or new.endswith('.csv'):
                self.data_export_button.disabled = False
            else:
                self.data_export_button.disabled = True
        def _export_figure(*args):
            if not self.export_file_input.value:
                raise RuntimeError('no output file defined')
            self.export_figure(self.export_file_input.value)
        def _export_data(*args):
            if not self.export_file_input.value:
                raise RuntimeError('no output file defined')
            self.export_data(self.export_file_input.value)
        self.export_file_input.on_change('value', _update_buttons)
        self.figure_export_button.on_click(_export_figure)
        self.data_export_button.on_click(_export_data)
        layout = column(self.export_file_input, self.figure_export_button, self.data_export_button)
        if not self.show_side_panel:
            layout.visible = False
        return layout
    def set_export_status(self, what, status):
        if what == 'figure':
            obj = self.figure_export_button
        elif what == 'data':
            obj = self.data_export_button
        else:
            raise ValueError("'{}' not supported; accepted values are: 'figure', 'data'".format(what))
        obj.label = '{} [{}]'.format(obj.label, status)
    def unset_export_status(self, what):
        if what == 'figure':
            self.figure_export_button.label = 'Export figure'
        elif what == 'data':
            self.data_export_button.label = 'Export data'
        else:
            raise ValueError("'{}' not supported; accepted values are: 'figure', 'data'".format(what))
    def export_figure(self, output_file):
        """
        Requires a working *selenium* driver (https://www.selenium.dev/selenium/docs/api/py/).

        Related attributes are `selenium_webdriver`, `figure_export_width` and `figure_export_height`.
        """
        export_kwargs = {}
        if self.selenium_webdriver is not None:
            try:
                from importlib import import_module
                options = import_module(self.selenium_webdriver.__module__[:-9]+'options')
                options = options.Options()
                options.headless = True
                webdriver = self.selenium_webdriver(options=options)
            except (ImportError, AttributeError):
                import selenium
                if self.selenium_webdriver in (selenium.webdriver.Safari, selenium.webdriver.Edge):
                    pass
                else:
                    import warnings, traceback
                    warnings.warn('could not access the webdriver''s options:\n'+traceback.format_exc(), ImportWarning)
                webdriver = self.selenium_driver()
            export_kwargs['webdriver'] = webdriver
        if self.figure_export_width is not None:
            export_kwargs['width'] = self.figure_export_width
        if self.figure_export_height is not None:
            export_kwargs['height'] = self.figure_export_height
        #self.set_export_status('figure', 'exporting...')
        try:
            doc = row(self.main_figure, self.colorbar_figure)
            from bokeh.io import export
            if output_file.endswith('.png'):
                export.export_png(doc, filename=output_file, **export_kwargs)
            elif output_file.endswith('.svg'):
                export.export_svg(doc, filename=output_file, **export_kwargs)
            else:
                raise NotImplementedError("format '{}' not supported".format(os.path.splitext(output_file)[1]))
            self.set_export_status('figure', 'done')
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            traceback.print_exc()
            self.set_export_status('figure', 'failed')
        #finally:
        #    self.unset_export_status('figure')
    def export_data(self, output_file):
        #self.set_export_status('data', 'exporting...')
        try:
            xy = self.model.current_sampling.tessellation
            try:
                xy = xy.spatial_mesh
            except AttributeError:
                pass
            xy = xy.cell_centers
            df = self.model.current_mapping.maps.copy()
            index = list(df.index)
            df = df.join(pd.DataFrame(xy[index], index=index, columns=['center x', 'center y']))
            df.to_csv(output_file, sep='\t')
            self.set_export_status('data', 'done')
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            traceback.print_exc()
            self.set_export_status('data', 'failed')
        #finally:
        #    self.unset_export_status('data')

        
def browse_maps(analyzer, **kwargs):
    """
    Runs a bokeh application for viewing the inferred parameter maps available through the analyzer.

    If ``bokeh serve`` is not called explicitly, then `browse_maps` opens a new tab browser.

    See also :class:`~tramway.analyzer.browser.Browser`.
    """
    model = Model(analyzer)
    controller = Controller(model, **kwargs)

    curdoc().add_root(controller.make_main_view())
    curdoc().title = 'TRamWAy viewer'

    if not model.spt_data_sources[1:]:
        controller.load_source(model.spt_data_sources[0])

    if not sys.argv[0].endswith('bokeh'):
        if analyzer.env.script and not \
            (os.path.isabs(analyzer.env.script) and \
                sys.argv[0] == os.path.basename(analyzer.env.script)):
            print('running bokeh server...\n')
            import subprocess
            p = subprocess.Popen([sys.executable, '-m',
                    'bokeh', 'serve', '--show', analyzer.env.script],
                cwd=os.getcwd(),
                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                close_fds=True,
                encoding='utf-8')
            try:
                out, err = p.communicate()
            except KeyboardInterrupt:
                out, err = p.communicate()
            #out = out.decode('utf-8')
            for line in out.splitlines():
                if not line:
                    continue
                print(line)
            if err:
                #print(err.decode('utf-8'))
                print(err)
            print('bokeh server shut down')
        else:
            import __main__ as main
            if not hasattr(main, '__file__'):
                print('''bokeh map browser cannot run in an interpreter
without the script attribute set''')




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from .abc import *


class UNetLocalizer(AnalyzerNode):
    """ Loads the weights of a U-Net network for image deconvolution.

    Not implemented yet, as long as the :mod:`~tramway.localization.UNet` package is broken.
    """
    __slots__ = ('_weights_locator',)
    def __init__(self, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        self._weights_locator = 0
    @property
    def weights_locator(self):
        """ *str*: Uniform resource locator for the U-Net weights """
        return self._weights_locator
    @weights_locator.setter
    def weights_locator(self, url):
        self._weights_locator = url
    def localize(self, stack):
        """
        Not implemented yet, as long as the :mod:`~tramway.localization.UNet` package is broken.
        """
        raise NotImplementedError

Localizer.register(UNetLocalizer)


class LocalizerInitializer(Initializer):
    """ Initializer class for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.localizer` main attribute.

    The :attr:`~tramway.analyzer.RWAnalyzer.localizer` attribute self-modifies
    on calling any of the *from_...* methods.

    """
    __slots__ = ()
    def from_UNet(self):
        """ Loads a trained U-Net network for image deconvolution.
        
        See also :class:`UNetLocalizer`."""
        self.specialize( UNetLocalizer )
    def from_unet(self):
        """ Alias for :meth:`from_UNet`. """
        self.from_UNet()


__all__ = [ 'Localizer', 'LocalizerInitializer', 'UNetLocalizer' ]





from . import *
from . import abc
from . import proxy
from . import proxied
from . import plugin
from . import post
from ..attribute import initializer_method

from_plugin   = initializer_method( TessellerInitializer.from_plugin   )
from_callable = initializer_method( TessellerInitializer.from_callable )




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from abc import *


class JobStep(metaclass=ABCMeta):
    """ Job step data.

    A job step object contains all the necessary input data for a job step
    to be performed as well as the output data resulting from the step completion.

    A job step object merely contains a reference to a shared workspace.

    The `resource_id` attribute refers to a series of the job steps that operate
    on the same subset of resource items.

    """
    @abstractmethod
    def get_workspace(self):
        raise NotImplementedError('abstract method')
    @abstractmethod
    def set_workspace(self, ws):
        raise NotImplementedError('abstract method')
    @abstractmethod
    def unset_workspace(self):
        raise NotImplementedError('abstract method')
    @property
    @abstractmethod
    def workspace_set(self):
        """ `bool`: Is the workspace set?"""
        raise NotImplementedError('abstract method')
    @property
    @abstractmethod
    def resource_id(self):
        """ `int`: Resource-related job ID.

        A job step is one of many that operate on a same subset of resources.
        `resource_id` uniquely designates this specific subset of resources.
        """
        raise NotImplementedError('abstract ro property')
    @property
    @abstractmethod
    def resources(self):
        """ `sequence`: Indices or keys of the required items of resource in the workspace.

        May be implemented as follows::

            return self.get_workspace().resources(self)

        """
        raise NotImplementedError('abstract ro property')


class Workspace(metaclass=ABCMeta):
    """ Parameter singleton.

    A computation will typically instanciates a unique workspace that will be
    embedded in and shared between multiple :class:`JobStep` instances.

    It embarks resources that a job step may access.
    """
    @abstractmethod
    def update(self, step):
        """ Update the workspace with a completed job step.

        Arguments:

            step (tramway.core.parallel.JobStep): completed job step.

        May be implemented as follows::

            step.set_workspace(self)

        """
        raise NotImplementedError('abstract method')

    @abstractmethod
    def __len__(self):
        """ `int`: Total number of resource items."""
        raise NotImplementedError('abstract method')

    @abstractmethod
    def resources(self, step):
        """ May be implemented as follows::

            return step.resources

        See also: :attr:`JobStep.resources`."""
        raise NotImplementedError('abstract method')


class WorkspaceExtension(metaclass=ABCMeta):
    """ See also: :class:`ExtendedWorkspace`."""
    @abstractmethod
    def pop_workspace_update(self):
        raise NotImplementedError('abstract method')
    @abstractmethod
    def push_workspace_update(self, upload):
        raise NotImplementedError('abstract method')


class ExtendedWorkspace(Workspace):
    """ Workspace that listens to external objects which implement :class:`WorkspaceExtension`
    provided that the job step data implements :class:`VehicleJobStep`.
    """

    @abstractmethod
    def push_extension_updates(self, updates):
        raise NotImplementedError('abstract method')

    @abstractmethod
    def pop_extension_updates(self, updates):
        raise NotImplementedError('abstract method')


class VehicleJobStep(metaclass=ABCMeta):
    """ Job step that keeps a reference to workspace-like objects that are not part
    of the main workspace.

    Changes to these external workspace-like objects are encoded together with the
    job step data so that it can be passed to and replayed in the other workers' workspaces.
    """
    @abstractmethod
    def pop_updates(self):
        raise NotImplementedError('abstract method')
    @abstractmethod
    def push_updates(self, updates):
        raise NotImplementedError('abstract method')


__all__ = [ 'JobStep', 'Workspace', 'WorkspaceExtension', 'ExtendedWorkspace', 'VehicleJobStep' ]




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .exception import *
from .base import *
import numpy as np
from scipy.sparse import csc_matrix, csr_matrix, bsr_matrix, dok_matrix, coo_matrix, dia_matrix


class DictGraph(Graph):
    """Default implementation of :class:`Graph`. Memory efficient and suitable for very dynamic graph
    with frequent modifications to its structure, but possibly computationally slow for :class:`Gas`.

    Consider using alternative implementations in module :mod:`.graph.alloc`."""
    def __init__(self, node_defaults, edge_defaults):
        Graph.__init__(self, node_defaults, edge_defaults)
        self.nodes = dict([ (attr, {}) for attr in node_defaults ])
        self._node_counter = 0
        self.neighbors = {}
        self.edges = dict([ (attr, {}) for attr in edge_defaults ])
        self._edge_counter = 0

    def connect(self, n1, n2, **kwargs):
        if self.has_node(n1) and self.has_node(n2):
            e = [ e for e, n in self.neighbors[n1] if n is n2 ]
            if e: # if nodes are already connected
                e = e[0]
            else:
                e = self._edge_counter
                self.neighbors[n1].append((e, n2))
                self.neighbors[n2].append((e, n1))
                self._edge_counter += 1
            for attr, default in self.edge_defaults.items():
                self.edges[attr][e] = kwargs.get(attr, default)
        else:   raise MissingNodeError((n1, n2))
        return e

    def disconnect(self, n1, n2, edge=None):
        if self.has_node(n1) and self.has_node(n2):
            e = edge
            if e is None:
                e = [ e for e, n in self.neighbors[n2] if n is n1]
                if e: e = e[0]
                else: e = None
            if e is None: # e can be 0
                #raise MissingEdgeError((n1, n2))
                pass
            else: # if nodes are connected
                self.neighbors[n1].remove((e, n2))
                self.neighbors[n2].remove((e, n1))
                for attr in self.edges:
                    del self.edges[attr][e]
        else:   raise MissingNodeError((n1, n2))

    def find_edge(self, n1, n2):
        if self.has_node(n1):
            e = [ e for e, n in self.neighbors[n1] if n is n2 ]
            if e: return e[0]
            else: return None
        else:
            return None

    def get_node_attr(self, n, attr):
        try:
            return self.nodes[attr][n]
        except KeyError:
            if attr in self.nodes:
                raise MissingNodeError(n)
            else:   raise NodeAttributeError(n)

    def set_node_attr(self, n, **kwargs):
        for attr, val in kwargs.items():
            if attr in self.nodes:
                if n in self.nodes[attr]:
                    self.nodes[attr][n] = val
                else:   raise MissingNodeError(n)
            else: raise NodeAttributeError(attr)

    def get_edge_attr(self, e, attr):
        try:
            return self.edges[attr][e]
        except KeyError:
            if attr in self.edges:
                raise MissingEdgeError(e)
            else:   raise EdgeAttributeError(attr)

    def set_edge_attr(self, e, **kwargs):
        for attr, val in kwargs.items():
            if attr in self.edges:
                if e in self.edges[attr]:
                    self.edges[attr][e] = val
                else:
                    raise MissingEdgeError(e)
            else:
                raise EdgeAttributeError(attr)

    def iter_edges(self):
        any_attr = list(self.edges.keys())[0]
        return self.edges[any_attr].keys()

    def iter_edges_from(self, n):
        try:
            return self.neighbors[n]
        except KeyError:
            raise MissingNodeError(n)

    def has_node(self, n):
        return n in self.neighbors

    def add_node(self, **kwargs):
        n = self._node_counter
        for attr, default in self.node_defaults.items():
            self.nodes[attr][n] = kwargs.get(attr, default)
        self.neighbors[n] = []
        self._node_counter += 1
        return n

    def del_node(self, n):
        if self.has_node(n):
            for edge, neighbor in list(self.neighbors[n]):
                self.disconnect(n, neighbor, edge)
            del self.neighbors[n]
            for attr in self.nodes:
                del self.nodes[attr][n]
        else:   raise MissingNodeError(n)

    @property
    def size(self):
        return len(self.neighbors)

    def iter_nodes(self):
        return self.neighbors.keys()

    def export(self, sparse='csr'):
        matrix = dict(csc=csc_matrix, csr=csr_matrix, bsr=bsr_matrix, coo=coo_matrix, \
            dok=lambda x: coo_matrix(x).todok(), \
            dia=lambda x: coo_matrix(x).todia())
            #lil=lambda x: coo_matrix(x).tolil(), \ # lil does not support explicit zeros
        any_attr = list(self.node_defaults.keys())[0]
        node_order = list(self.nodes[any_attr].keys())
        node_map = np.zeros(self._node_counter)
        node_map[node_order] = np.arange(0, len(node_order))
        any_attr = list(self.edge_defaults.keys())[0]
        edge_order = list(self.edges[any_attr].keys())
        edge_map = np.zeros(self._edge_counter)
        edge_map[edge_order] = np.arange(0, len(edge_order))
        edges, nodes1, nodes2 = zip(*[ (e, n1, n2) for n1, ns in self.neighbors.items() \
                                for e, n2 in ns ])
        edges = edge_map[np.asarray(edges)]
        nodes1 = node_map[np.asarray(nodes1)]
        nodes2 = node_map[np.asarray(nodes2)]
        A = matrix[sparse]((edges, (nodes1, nodes2)))
        V = {}
        for attr, data in self.nodes.items():
            V[attr] = np.asarray(list(data.values()))
        E = {}
        for attr, data in self.edges.items():
            E[attr] = np.asarray(list(data.values()))
        return (A, V, E)

    def square_distance(self, attr, eta, eta2=None):
        if eta2 is None:
            eta2 = np.dot(eta, eta)
        x = np.vstack(self.nodes[attr].values())
        d = np.sum(x * x, axis=1) + eta2 - 2.0 * np.dot(x, eta)
        n = list(self.iter_nodes())
        def index_to_node(i):
            if isinstance(i, int):
                return n[i]
            else:   return [ n[j] for j in i ]
        return (d, index_to_node)




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from . import rc
from .rc import *
from .lazy import ro_property_assert, Lazy # hide PermissionError
from .namedcolumns import *
from .chain import *
from . import exceptions
from .analyses import *
from .scaler import *
from .xyt import *
#from .hdf5 import *




# -*- coding: utf-8 -*-

# Copyright © 2018-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .base import *
import tramway.inference.gradient as grad
import numpy as np
import pandas as pd
import scipy.sparse as sparse


class DynamicTranslocations(Translocations):
    __slots__ = ('center_t',)
    def __init__(self, index, translocations, center=None, span=None, boundary=None):
        Translocations.__init__(self, index, translocations, center, span, boundary)
        rt = self.center
        if isinstance(self.data, pd.DataFrame):
            t_pos = list(self.data.columns).index(self.time_col)
        else:
            t_pos = self.time_col
        try:
            self.center_t = rt[t_pos].tolist()
        except IndexError:
            raise RuntimeError('missing time column; was the tessellation made with defined time_dimension?')
        others = np.ones(rt.size, dtype=bool)
        others[t_pos] = False
        self.center_r = rt[others]

    @property
    def center_r(self):
        return self.center
    @center_r.setter
    def center_r(self, r):
        self.center = r


class DynamicCells(Distributed):

    def __init__(self, cells, adjacency, index=None, center=None, span=None, central=None,
        boundary=None, spatial_adjacency=None, temporal_adjacency=None, time_adjacency=None):
        Distributed.__init__(self, cells, adjacency, index, center, span, central, boundary)
        if time_adjacency is None:
            time_adjacency = temporal_adjacency
        if time_adjacency is None and spatial_adjacency is None and \
                adjacency.dtype not in (bool, np.bool_):
            # separate spatial adjacency and temporal adjacency
            A = adjacency.tocoo()
            ok = 1 < A.data
            time_adjacency = sparse.csr_matrix((np.ones(np.sum(ok), dtype=bool), (A.row[ok], A.col[ok])), shape=A.shape)
            ok = A.data == 1
            spatial_adjacency = sparse.csr_matrix((np.ones(np.sum(ok), dtype=bool), (A.row[ok], A.col[ok])), shape=A.shape)
        self.spatial_adjacency = spatial_adjacency
        self.time_adjacency = time_adjacency

    @property
    def temporal_adjacency(self):
        """
        alias for :attr:`time_adjacency`.
        """
        return self.time_adjacency

    def neighbours(self, i):
        """
        Indices of spatial-neighbour cells.

        Argument:

            i (int): cell index.

        Returns:

            numpy.ndarray: indices of the neighbour cells of cell *i*.

        """
        return self.spatial_adjacency.indices[self.spatial_adjacency.indptr[i]:self.spatial_adjacency.indptr[i+1]]

    def time_neighbours(self, i):
        """
        Indices of time-neighbour cells.

        Argument:

            i (int): cell index.

        Returns:

            numpy.ndarray: indices of the neighbour cells of cell *i*.

        """
        return self.time_adjacency.indices[self.time_adjacency.indptr[i]:self.time_adjacency.indptr[i+1]]

    def time_derivative(self, i, X, index_map=None, na=np.nan, **kwargs):
        cell = self.cells[i]
        t0 = cell.center_t

        # cache neighbours (indices and center locations)
        if not isinstance(cell.cache, dict):
            cell.cache = {}
        try:
            i, adjacent, t = cell.cache['time_derivative']
        except KeyError:
            A = self.time_adjacency
            adjacent = _adjacent = A.indices[A.indptr[i]:A.indptr[i+1]]
            if index_map is not None:
                adjacent = index_map[_adjacent]
                ok = 0 <= adjacent
                assert np.all(ok)
                #adjacent, _adjacent = adjacent[ok], _adjacent[ok]
            if _adjacent.size:
                t = np.array([ self.cells[j].center_t for j in _adjacent ])
                before, after = t<t0, t0<t

                # pre-compute the "t" term
                u, v = before, after
                if not np.any(u):
                    u = None
                if not np.any(v):
                    v = None

                if u is None:
                    if v is None:
                        t = None
                    else:
                        t = 1. / (t0 - np.mean(t[v]))
                elif v is None:
                    t = 1. / (t0 - np.mean(t[u]))
                else:
                    t = np.r_[t0, np.mean(t[u]), np.mean(t[v])]

                if t is not None:
                    t = (u, v, t)

            else:
                t = None

            if index_map is not None:
                i = index_map[i]
            cell.cache['time_derivative'] = (i, adjacent, t)

        if t is None:
            return None

        x0, x = X[i], X[adjacent]

        # compute the derivative
        u, v, t = t
        #u, v, t= before, after, t term
        if u is None:
            if v is None:
                deriv = na
            else:
                # 1./X = X0[j] - np.mean(X[v,j])
                deriv = (x0 - np.mean(x[v])) * t
        elif v is None:
            deriv = (x0 - np.mean(x[u])) * t
        else:
            deriv = grad._vander(t, np.r_[x0, np.mean(x[u]), np.mean(x[v])])

        return deriv

    def time_variation(self, i, X, index_map=None, na=0., **kwargs):
        cell = self.cells[i]
        t0 = cell.center_t

        # cache neighbours (indices and center locations)
        if not isinstance(cell.cache, dict):
            cell.cache = {}
        try:
            i, adjacent, t = cell.cache['time_derivative']
        except KeyError:
            A = self.time_adjacency
            adjacent = _adjacent = A.indices[A.indptr[i]:A.indptr[i+1]]
            if index_map is not None:
                adjacent = index_map[_adjacent]
                ok = 0 <= adjacent
                assert np.all(ok)
                #adjacent, _adjacent = adjacent[ok], _adjacent[ok]
            if _adjacent.size:
                t = np.array([ self.cells[j].center_t for j in _adjacent ])
                before, after = t<t0, t0<t

                # pre-compute the "t" term
                u, v = before, after
                if not np.any(u):
                    u = None
                if not np.any(v):
                    v = None

                if u is None:
                    if v is None:
                        t = None
                    else:
                        t = 1. / (t0 - np.mean(t[v]))
                elif v is None:
                    t = 1. / (t0 - np.mean(t[u]))
                else:
                    t = np.r_[t0, np.mean(t[u]), np.mean(t[v])]

                if t is not None:
                    t = (u, v, t)

            else:
                t = None

            if index_map is not None:
                i = index_map[i]
            cell.cache['time_derivative'] = (i, adjacent, t)

        if t is None:
            return None

        x0, x = X[i], X[adjacent]

        # compute the derivative
        u, v, t = t
        #u, v, t= before, after, t term
        if u is None:
            if v is None:
                delta = np.r_[na, na]
            else:
                # 1./X = X0[j] - np.mean(X[v,j])
                delta = np.r_[na, (x0 - np.mean(x[v])) * t]
        elif v is None:
            delta = np.r_[(x0 - np.mean(x[u])) * t, na]
        else:
            t0, tu, tv = t
            delta = np.r_[
                (x0 - np.mean(x[u])) / (t0 - tu),
                (x0 - np.mean(x[v])) / (t0 - tv),
                ]
            #delta = np.abs(delta)

        return delta

    def temporal_variation(self, *args, **kwargs):
        """
        alias for :meth:`time_variation`.
        """
        return self.time_variation(*args, **kwargs)

    def spatial_variation(self, *args, **kwargs):
        """
        alias for :meth:`local_variation`.
        """
        return self.local_variation(*args, **kwargs)


__all__ = ['DynamicTranslocations', 'DynamicCells']




#!/usr/bin/env python


from tramway.core.xyt import load_xyt, crop
import numpy as np


def cropping_utility():
    import argparse
    parser = argparse.ArgumentParser(prog='crop',
        description='crop 2D trajectories')
    parser.add_argument('--columns', help="input trajectory file column names (comma-separated list)")
    parser.add_argument('-q', '--quiet', action='store_true', help="do not ask to overwrite the file")
    parser.add_argument('input_file', help='path to input trajectory file')
    parser.add_argument('bounding_box', help='bounding box as left,bottom,right,top')
    parser.add_argument('output_file', nargs='?', help='path to output trajectory file')
    args = parser.parse_args()

    bounding_box = np.array([ float(x) for x in args.bounding_box.split(',') ])
    dim = int(round(float(bounding_box.size) * .5))
    bounding_box[dim:] -= bounding_box[:dim]

    load_kwargs = {}
    if args.columns:
        load_kwargs['columns'] = args.columns.split(',')
    trajectories = load_xyt(args.input_file, **load_kwargs)

    cropped_trajectories = crop(trajectories, bounding_box.tolist())

    if args.output_file:
        output_file = args.output_file
    else:
        output_file = args.input_file
        if not args.quiet:
            invite = "overwrite file '{}': [N/y] ".format(output_file)
            try:
                answer = raw_input(invite) # Py2
            except NameError:
                answer = input(invite)
            if not (answer and answer[0].lower() == 'y'):
                return

    cropped_trajectories.to_csv(output_file, sep='\t', index=False)


if __name__ == '__main__':
    cropping_utility()




# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Useful functions
"""

import glob
import pickle
import os

import tqdm
import pandas as pd


def extract_i(file):
    """Extracts the id of a file generated by the cluster."""
    return file.split('\\')[3].split('_')[2].split('.')[0]


def concat_job_files(job_name, output_dir, output_name, get_rws=True,
                     df_path='Y:\data\df', dict_path='Y:\data\dict',
                     rw_path='Y:\data\rws'):
    """Concatenates and deletes files produced by single jobs when doing a
    batch array of jobs.
    """
    dfs = []
    prms = {}
    rws = []
    rws_paths = glob.glob(f'{rw_path}\df_{job_name}_*')
    df_paths = glob.glob(f'{df_path}\df_{job_name}_*')
    dict_paths = glob.glob(f'{dict_path}\dict_{job_name}_*')
    for rw_path, df_path, dict_path in tqdm.tqdm_notebook(
            zip(rws_paths, df_paths, dict_paths), total=len(df_paths)):
        try:
            if get_rws:
                rws.append(pd.read_csv(rw_path, index_col=0))
                os.remove(rw_path)
            dfs.append(pd.read_csv(df_path, index_col=0))
            with open(dict_path, 'rb') as handle:
                dict_i = pickle.load(handle)
            prms = {**prms, **dict_i}
            os.remove(df_path)
            os.remove(dict_path)
        except:
            print(f'Could not add file {extract_i(df_path)}')
    df = pd.concat(dfs, sort=True).reset_index()
    if get_rws:
        rws = pd.concat(rws).reset_index()
        rws.to_feather(f'{output_dir}\\RWs_{output_name}.feather')
    df.to_feather(f'{output_dir}\\features_{output_name}.feather')
    with open(f'{output_dir}\\prms_{output_name}.pickle', 'wb') as f:
        pickle.dump(prms, f)


def load(Dir, name, trajs=True):
    """Loads files associated with a job array (of producing random walk
    features). Returns parameters of the generated random walks, a DataFrame,
    and possibly raw trajectories.
    """
    path = f'{Dir}\\features_{name}.feather'
    df_feat = pd.read_feather(path)
    with open(f'{Dir}\\prms_{name}.pickle', 'rb') as f:
        prms = pickle.load(f)
    if trajs:
        RWs = pd.read_feather(f'{Dir}\\RWs_{name}.feather')
        return prms, df_feat, RWs
    else:
        return prms, df_feat



# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute.abc import Attribute, abstractmethod

class Time(Attribute):
    """
    Abstract base class for the :attr:`~tramway.analyzer.RWAnalyzer.time` attribute
    of an :class:`~tramway.analyzer.RWAnalyzer` object.
    """
    @abstractmethod
    def segment(self, spt_dataframe, tessellation=None):
        """
        Segments the SPT data, combines the segmentation with a spatial tessellation if any,
        and returns a :class:`~tramway.tessellation.time.TimeLattice` object.
        """
        pass
    @abstractmethod
    def as_time_segments(self, sampling, maps=None, index=None, return_index=False, return_times=True):
        """ Generator function; yields single-segment sampling and map objects.

        Slices `sampling` and `maps` and yields tuples with elements in the following order:

        * segment index (*int*), if `return_index` is :const:`True`,
        * segment start and stop times *(float, float)*, if `return_times` is :const:`True` (default),
        * segment :class:`~tramway.tessellation.base.Partition` object, from `sampling`,
        * segment maps (*pandas.DataFrame*) from `maps`, if `maps` is defined.

        `index` is a selector on the segment index, either as an *int*, a *set* of *int*,
        a *sequence* of *int*, or a *callable* that takes a segment index (*int*) as input argument
        and returns a *bool*.
        """
        pass
    @property
    @abstractmethod
    def dt(self):
        """
        *float*: See :attr:`~tramway.analyzer.spt_data.SPTParameters.dt`
        """
        pass
    @dt.setter
    @abstractmethod
    def dt(self, dt):
        pass

